////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MKB;

import java.util.*;
import DSRG.VSA.VSA;
import DSRG.VSA.Schema_Change;

/**
 *
 * one relation (consists of attributes).<BR>
 * This is meta-data (MKB) only.
 * @author koeller 111199
 *
 */
public class Relation {
    
    /**
     * the logical name of this relation
     */
    private String relName=null;
    
    /** a reference to the parent IS object */
    IS parent=null;
    
    /** The attributes of this relation. <BR> The elements are of type
     * "EVEDemo.MKB.Attribute" */
    private Vector attributes = new Vector();
    
    /**
     * the (references) to joins in which this relation takes part */
    private Vector joins = new Vector();
    
    /**
     * the (references) to PCs in which this relation takes part */
    private Vector PCs = new Vector();
    
    /**
     * Size of this relation ( number of tuples ). Used for QC model.
     *
     */
    private int size = 0;
    
    /**
     * Constructor used By MKBParser. It has one more parameter: size
     * (# of tuples) */

    public Relation (IS myIS, String myName, int size){
        parent = myIS;
        relName = myName;
        this.size = size;
        if (!parent.addRelation(this))
            System.err.println("MKB.Relation: Relation"+myIS+"."+myName+" could not be added to MKB.");
    }

    /** gives a string representation of this Relation */
    public String displayAll() {
        String returnString=new String(relName);
        for(int i=0;i<attributes.size();i++)
            returnString=returnString+" "+
                ((Attribute)attributes.elementAt(i)).displayAll();
        return returnString;
    }
    
    /**
     * Dump the Relation object.
     */
    public String toString() {
        return relName + ":" + attributes;
    }
    
    /**
     * returns the vector of attributes in this relation
     */
    public Vector getAttributes(){
        return attributes;
    }
    
    /**
     * returns the vector of join constraints in this relation
     */
    public Vector getJCs(){
        return joins;
    }
    
    /**
     * returns the vector of PC constraints in this relation
     */
    public Vector getPCs(){
        return PCs;
    }

    /**
     * returns the vector of Type constraints in this relation. <BR>
     * Not sure if needed.  */
    public Vector getTCs() {
        Enumeration attrEnum = attributes.elements();
        Vector returnVec = new Vector();
        Attribute myAttr = null;
        while (attrEnum.hasMoreElements()) {
            myAttr = (Attribute)attrEnum.nextElement();
            returnVec.addElement(myAttr.getTypeConstraint());
        }
        return returnVec;
        
    }

    /** returns the number of tuples as stated in the MKB file
        (QC-Model system parameter) */

    public int getRelationSize() {
        return size;
    }
    
    /**
     * finds an attribute object by name */
    
    public Attribute findAttribute(String Name){
        if (attributes == null) return null;
        for(int i=0;i<attributes.size();i++){
            if(Name.equals(((Attribute)attributes.elementAt(i)).getName())){
                return (Attribute)attributes.elementAt(i);
            }
        }
        return null;
    }
    
    /** add existing join constraint object to the join list of this
        attribute */
    
    public boolean addJoinConstraint(JoinConstraint theJC){
        joins.addElement(theJC);
        return true;
    }
    
    /** remove join constraint from this attribute's list (only the
        reference, of course)*/
    public boolean deleteJoinConstraint(JoinConstraint theJC){
        return joins.removeElement(theJC);
    }
    

    /** add existing join constraint object to the join list of this
        attribute */
    
    public boolean addPCConstraint(PCConstraint thePC){
        PCs.addElement(thePC);
        return true;
    }
    
    /** remove PC constraint from this attribute's list (only the
        reference, of course)*/
    public boolean deletePCConstraint(PCConstraint thePC){
        return PCs.removeElement(thePC);
    }
    
    /**
     * returns the name of this relation
     */
    public String getName(){
        return relName;
    }
    
    
    /**
     * get a string of the form ISName.RelName */
    public String getFullName(){
        return new String(parent.getName()+"."+relName);
    }
    
    /** return handle to this relation's IS */
    
    public IS getIS() {
        return parent;
    }
    
    /** return name of parent IS */
    public String getISName() {
        return parent.getName();
    }
    
    /**
     * Add an Attribute Object to this Relation. Called from MKB Parser
     * during system initialization.
     * @author Xin Zhang.  */
    public boolean addAttribute (Attribute attr){
        if(findAttribute(attr.getName())!=null) return false;
        attributes.addElement(attr);
        return true;
    }
    
    /**
     * Add an Attribute Object to this Relation. Called from MKB Parser
     * during system initialization.
     * @author Xin Zhang.  */
    public boolean deleteAttribute (Attribute attr){
        return attributes.removeElement(attr);
    }

    /**
     * renames MKB relation object only.  */
    
    boolean changeRelationName(String newName) {
        relName = new String(newName);
        return true;
    }
    
//=================================================================
//=================================================================
    // for debugging
//=================================================================
    /** function to be used for debugging purposes only */

 private void displayRel() {

    Attribute a1 = null;  
    Vector atts = this.getAttributes(); // we need these later

   int width = 10;

    // 3.1. first the header

   for (int i=0;i<atts.size();i++) {
      a1 = (Attribute)atts.elementAt(i);
      System.out.print(pad(a1.getName(),width));  // pad() is defined below, it's easier...
   }
   System.out.println();
    
   for (int i=0;i<atts.size();i++) {
      a1 = (Attribute)atts.elementAt(i);
      System.out.print(pad(a1.getName(),width));  // pad() is defined below, it's easier...
   }
   System.out.println();

   for (int i=0;i<atts.size();i++) {
      a1 = (Attribute)atts.elementAt(i);
      System.out.print(pad(a1.getType(),width));  // pad() is defined below, it's easier...
   }
   System.out.println();

   for (int i=0;i<atts.size();i++) {
      a1 = (Attribute)atts.elementAt(i);
      System.out.print(pad((new Integer(a1.getSize()).toString()),width));  // pad() is defined below, it's easier...
   }
   System.out.println();

 }
   
    /** a private function for output formatting. Use only for debugging.  */
    
    private static String pad(String s, int length) {
        
        String pad=s;
        if (s.length()<length) {
            for (int i=0; i<length-s.length();i++)
          pad=pad+" ";
        }
        else pad=s.substring(0,s.length()-1)+" ";
        return pad;
        
    }
}

