////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MKB.visitor;

import DSRG.MKB.*;
import DSRG.MKB.syntaxtree.*;


import java.util.*;

/**
 * Construct a MKB object from AST.
 */
public class MKBVisitor extends DepthFirstVisitor {

private static final boolean DEBUG = false;

   
   MKB mkb = null;

   //
   // Auto class visitors--probably don't need to be overridden.
   //
   public void visit(NodeList n) {
      for ( Enumeration e = n.elements(); e.hasMoreElements(); )
         ((Node)e.nextElement()).accept(this);
   }

   public void visit(NodeListOptional n) {
      if ( n.present() )
         for ( Enumeration e = n.elements(); e.hasMoreElements(); )
            ((Node)e.nextElement()).accept(this);
   }

   public void visit(NodeOptional n) {
      if ( n.present() )
         n.node.accept(this);
   }

   public void visit(NodeSequence n) {
      for ( Enumeration e = n.elements(); e.hasMoreElements(); )
         ((Node)e.nextElement()).accept(this);
   }

   public void visit(NodeToken n) { }

   //
   // User-generated visitor methods below
   //

  /*
   * I am going to create a IS object while we visit the IS-Definition node.
   * I think the visitor design should let us pass a parameter like: object.
   * Then, we can have more flexibility. But, ok, I will use a global variable
   * inside this parser to share the IS object.
   */
  String curISName; // current IS name.
  String curRName; // current Relation name.
  String curAName; // current Attribute name.
  String curDBType; // current DB Type.
  String curDBURL; // current DB URL.
  String curUsername; // current Username.
  String curPassword; // current Password.
    String curBOperator; // boolean operator
    String curValue; // values

  int curBF; // current Blocking Factor. 

  int curRSize; // current relation size. ( # of tuples)

  int curASize; // current Attribute Size. ( In Bytes )

  String curType; // current attribute type, e.g. VARCHAR(20), it's VARCHAR.
  int curTypeSize; // current type size, e.g. VARCHAR(20), it's 20.

   /**
    * f0 -> schema()
    * f1 -> <SEMICOLON>
    * f2 -> ( constraint() <SEMICOLON> )*
    * f3 -> ( local_condition() <SEMICOLON> )*
    */
   public void visit(MKBFile n) {
     mkb = new MKB();
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
   }

   /**
    * f0 -> <SCHEMA>
    * f1 -> <LPAREN>
    * f2 -> IS_definition()
    * f3 -> ( <COMMA> IS_definition() )*
    * f4 -> <RPAREN>
    */
   public void visit(schema n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);
   }



/**
 * New Grammar production:
 * f0 -> ISName()
 * f1 -> <LPAREN>
 * f2 -> dbtype()
 * f3 -> <COLON>
 * f4 -> <LPAREN>
 * f5 -> DBURL()
 * f6 -> <COMMA>
 * f7 -> username()
 * f8 -> <COMMA>
 * f9 -> password()
 * f10 -> [ <COMMA> <BF> <EQUAL> blocking_factor() ]
 * f11 -> <RPAREN>
 * f12 -> [ <COMMA> relation_definition() ( <COMMA> relation_definition() )* ]
 * f13 -> <RPAREN>
 */
   public void visit(IS_definition n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);
      n.f5.accept(this);
      n.f6.accept(this);
      n.f7.accept(this);
      n.f8.accept(this);
      n.f9.accept(this);
      n.f10.accept(this);
      n.f11.accept(this);
      /* algorithm:
	 .search mkb for specific IS
	 .if we cannot find it, then add a new IS.
       */
      if (DEBUG) System.err.println("MKBVisitor::visit: Adding IS "+curISName);
      if(mkb.findIS(curISName)==null){
	IS newIS = new IS(mkb,curISName, curDBType, 
			  curDBURL.substring(1, curDBURL.length()-1), 
			  curUsername.substring(1, curUsername.length()-1), 
			  curPassword.substring(1, curPassword.length()-1), 
			  curBF,curISName.equals("LOCAL_IS"));
      }
      n.f12.accept(this);
      n.f13.accept(this);
   }

   /**
    * f0 -> <STRING>
    */
   public void visit(DBURL n) {
      n.f0.accept(this);
      curDBURL = n.f0.toString();
   }

   /**
    * f0 -> <LPAREN>
    * f1 -> RelationName()
    * f2 -> [ <LPAREN> <SIZE> <EQUAL> relation_size() <RPAREN> ]
    * f3 -> <COLON>
    * f4 -> attribute_definition()
    * f5 -> ( <COMMA> attribute_definition() )*
    * f6 -> <RPAREN>
    */
   public void visit(relation_definition n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
      /* algorithm:
	 .search IS for specific Relation
	 .if we cannot find it, then add a new Relation.
      */
      IS the_IS = mkb.findIS(curISName);
      if(the_IS.findRelation(curRName) == null){
	//System.out.println("Relation: " + curRName +  
	//		   " is added to IS: " + curISName);
	Relation rel = new Relation(the_IS, curRName, curRSize);
        // relation adds itself automagically to its IS!
      }
      n.f4.accept(this);
      n.f5.accept(this);
      n.f6.accept(this);
   }


   /**
    * f0 -> AttributeName()
    * f1 -> [ <LPAREN> <SIZE> <EQUAL> attribute_size() <RPAREN> ]
    */
   public void visit(attribute_definition n) {
      n.f0.accept(this);
      n.f1.accept(this);
      /*
	algorithm: add the attribute to current relation.
       */
      IS the_IS = mkb.findIS(curISName);
      Relation the_R = the_IS.findRelation(curRName);
      if(the_R.findAttribute(curAName) == null){
	Attribute attr = new Attribute(the_R, curAName, curASize);
        // the attribute adds itself automagically to its relation
      }
   }

   /**
    * f0 -> type_constraint()
    *       | join_constraint()
    *       | pc_constraint()
    */
   public void visit(constraint n) {
      n.f0.accept(this);
   }

    TypeConstraint curTC = null;
    String RName = null;
    String ISName = null;
    String AName = null;

    /**
    * f0 -> <TC>
    * f1 -> <LPAREN>
    * f2 -> ISName()
    * f3 -> <DOT>
    * f4 -> RelationName()
    * f5 -> <COLON>
    * f6 -> AttributeList()
    * f7 -> <RPAREN>
    */
   public void visit(type_constraint n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      ISName = curISName;
      n.f3.accept(this);
      n.f4.accept(this);
      RName = curRName;
      n.f5.accept(this);
      n.f6.accept(this);
      n.f7.accept(this);
   }

   /**
    * f0 -> AttributeName()
    * f1 -> <IN>
    * f2 -> type()
    * f3 -> [ <COMMA> AttributeList() ]
    */
   public void visit(AttributeList n) {
       // String TCType = curType;
      n.f0.accept(this);
      AName = curAName;
      n.f1.accept(this);
      n.f2.accept(this);
      Attribute A1 = mkb.findAttribute(ISName, RName, AName);
      //System.out.println(ISName+"."+RName+"."+AName+"("+curType+":"+curTypeSize+")");
      A1.setType(curType);
      // if no size was assigned in the Schema, use the type size.
      if (A1.getSize()==-1) 
          A1.setSize(curTypeSize);
      //      if (curTypeSize!=-1)
      //    TCType += "("+new Integer(curTypeSize).toString()+")";
      // new TypeConstraint(AName, TCType);

      n.f3.accept(this);
   }

  // Used for creating the Join Constraint.
  JoinConstraint curJC = null; // current Join Constraint
  String ISName1 = null;
  String RName1 = null;
  String ISName2 = null;
  String RName2 = null;
  String JCPC = null; // used to distinguish the attribute 
                      // pair is for PC or for JC.
                      // if "JC" then, it's for JC,
                      // if "PC" then, it's for PC.
  // too bad.. programming ...


   /**
    * f0 -> <JC>
    * f1 -> <LPAREN>
    * f2 -> ISName()
    * f3 -> <DOT>
    * f4 -> RelationName()
    * f5 -> <JOIN>
    * f6 -> ISName()
    * f7 -> <DOT>
    * f8 -> RelationName()
    * f9 -> [ <LPAREN> <JS> <EQUAL> join_selectivity() <RPAREN> ]
    * f10 -> <COLON>
    * f11 -> attribute_pair()
    * f12 -> ( <COMMA> attribute_pair() )*
    * f13 -> <RPAREN>
    */
   public void visit(join_constraint n) {
     JCPC = "JC";
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      ISName1 = curISName;
      n.f3.accept(this);
      n.f4.accept(this);
      RName1 = curRName;
      n.f5.accept(this);
      n.f6.accept(this);
      ISName2 = curISName;
      n.f7.accept(this);
      n.f8.accept(this);
      RName2 = curRName;
      Relation R1, R2;
      if((R1 = mkb.findRelation(ISName1, RName1)) != null &&
         (R2 = mkb.findRelation(ISName2, RName2)) != null ){
	curJC = new JoinConstraint(R1, R2);
	R1.addJoinConstraint(curJC);
	R2.addJoinConstraint(curJC);
	mkb.addJoinConstraint(curJC);
      }else{
	System.out.println("JC erorr:" + ISName1 + "." + RName1
			   + " or " + ISName2 + "." + RName2 + 
			   " is not defined. ");
      }
      n.f9.accept(this);
      n.f10.accept(this);
      n.f11.accept(this);
      n.f12.accept(this);
      n.f13.accept(this);
   }

  // For PC Constraint we have following global variables.
  String curOperator = null;
  PCConstraint curPC = null;

   /**
    * f0 -> <PC>
    * f1 -> <LPAREN>
    * f2 -> ISName()
    * f3 -> <DOT>
    * f4 -> RelationName()
    * f5 -> operator()
    * f6 -> ISName()
    * f7 -> <DOT>
    * f8 -> RelationName()
    * f9 -> <COLON>
    * f10 -> attribute_pair()
    * f11 -> ( <COMMA> attribute_pair() )*
    * f12 -> <RPAREN>
    */
   public void visit(pc_constraint n) {
     JCPC = "PC";
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      ISName1 = curISName;
      n.f3.accept(this);
      n.f4.accept(this);
      RName1 = curRName;
      n.f5.accept(this);
      n.f6.accept(this);
      ISName2 = curISName;
      n.f7.accept(this);
      n.f8.accept(this);
      RName2 = curRName;
      Relation R1, R2;
      if((R1=mkb.findRelation(ISName1, RName1))!=null &&
	 (R2=mkb.findRelation(ISName2, RName2))!=null){
	curPC = new PCConstraint(R1, R2, curOperator);
	R1.addPCConstraint(curPC);
	R2.addPCConstraint(curPC);
	mkb.addPCConstraint(curPC);
      }else{
	System.out.println("PC error: "  + ISName1 + "." + RName1
			   + " or " + ISName2 + "." + RName2 + 
			   " is not defined. ");
      }
      n.f9.accept(this);
      n.f10.accept(this);
      n.f11.accept(this);
      n.f12.accept(this);
   }

    LocalCondition LC = null;
    String bOperator=null;
    String value = null;
    double curLCSelectivity=0.0;

   /**
    * f0 -> <LC>
    * f1 -> <LPAREN>
    * f2 -> ISName()
    * f3 -> <DOT>
    * f4 -> RelationName()
    * f5 -> <DOT>
    * f6 -> AttributeName()
    * f7 -> boolean_operator()
    * f8 -> value()
    * f9 -> <LPAREN>
    * f10 -> <SEL>
    * f11 -> <EQUAL>
    * f12 -> local_selectivity()
    * f13 -> <RPAREN>
    * f14 -> <RPAREN>
    */

   public void visit(local_condition n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      ISName =  curISName;
      n.f3.accept(this);
      n.f4.accept(this);
      RName = curRName;
      n.f5.accept(this);
      n.f6.accept(this);
      AName = curAName;
      n.f7.accept(this);
      bOperator = curBOperator;
      n.f8.accept(this);
      value = curValue;
      n.f9.accept(this);
      n.f10.accept(this);
      n.f11.accept(this);
      n.f12.accept(this);
      LC = new LocalCondition(curISName.trim()+"."+curRName.trim()+"."+curAName.trim(),
                              curBOperator,
                              curValue,
                              curLCSelectivity);
      mkb.addLocalCondition(LC);
      n.f13.accept(this);
      n.f14.accept(this);
   }

   /**
    * f0 -> <LPAREN>
    * f1 -> AttributeName()
    * f2 -> <COMMA>
    * f3 -> AttributeName()
    * f4 -> <RPAREN>
    */
   public void visit(attribute_pair n) {
     // We use curJC here.
      n.f0.accept(this);
      n.f1.accept(this);
      Attribute A1 = mkb.findAttribute(ISName1, RName1, curAName);
      n.f2.accept(this);
      n.f3.accept(this);
      Attribute A2 = mkb.findAttribute(ISName2, RName2, curAName);
      n.f4.accept(this);
      // I think we haven't includ the "extent" in the grammar, so I use "=".
      if(JCPC.equals("JC"))
	curJC.addCondition(A1, A2, "=");
      else if (JCPC.equals("PC"))
	curPC.addPair(A1, A2);
   }

   /**
    * f0 -> <LITERAL>
    */
   public void visit(ISName n) {
      n.f0.accept(this);
      curISName = n.f0.toString();
   }

   /**
    * f0 -> <LITERAL>
    */
   public void visit(RelationName n) {
      n.f0.accept(this);
      curRName = n.f0.toString();
   }

   /**
    * f0 -> <LITERAL>
    */
   public void visit(AttributeName n) {
      n.f0.accept(this);
      curAName = n.f0.toString();
   }

   /**
    * f0 -> <LITERAL>
    * f1 -> [ <LPAREN> typesize() <RPAREN> ]
    */
   public void visit(type n) {
      n.f0.accept(this);
      curType = n.f0.toString();
      curTypeSize = -1; // set default value of size.
      n.f1.accept(this);
   }

   /**
    * f0 -> <INTEGER>
    */
   public void visit(typesize n) {
      n.f0.accept(this);
      curTypeSize = (new Integer(n.f0.toString())).intValue();
   }



   /**
    * f0 -> <LITERAL>
    */
   public void visit(dbtype n) {
      n.f0.accept(this);
      curDBType = n.f0.toString();
   }

   /**
    * f0 -> <STRING>
    */
   public void visit(username n) {
      n.f0.accept(this);
      curUsername = n.f0.toString();
   }

   /**
    * f0 -> <STRING>
    */
   public void visit(password n) {
      n.f0.accept(this);
      curPassword = n.f0.toString();
   }

   /**
    * f0 -> <INTEGER>
    */
   public void visit(blocking_factor n) {
      n.f0.accept(this);
      curBF = (new Integer(n.f0.toString())).intValue();
   }

   /**
    * f0 -> <NOMORE>
    *       | <LESS>
    *       | <EQUAL>
    *       | <GREAT>
    *       | <NOLESS>
    */
   public void visit(boolean_operator n) {
      n.f0.accept(this);
      curBOperator =  ((NodeChoice)n.f0).choice.toString();
      // hacked by koeller 052199
   }

   /**
    * f0 -> <SUBSET>
    *       | <SUPERSET>
    *       | <EQUIV>
    *       | <APPROX>
    */
   public void visit(operator n) {
      n.f0.accept(this);
      curOperator = ((NodeChoice)n.f0).choice.toString();
   }

   /**
    * f0 -> <INTEGER>
    */
   public void visit(relation_size n) {
      n.f0.accept(this);
      curRSize = (new Integer(n.f0.toString())).intValue();
   }

   /**
    * f0 -> <INTEGER>
    */
   public void visit(attribute_size n) {
      n.f0.accept(this);
      curASize = (new Integer(n.f0.toString())).intValue();
   }

   /**
    * f0 -> <FLOAT>
    */
   public void visit(join_selectivity n) {
      n.f0.accept(this);
      curJC.setJoinSelectivity(new Double(n.f0.toString()).doubleValue());
   }

   /**
    * f0 -> <FLOAT>
    */
   public void visit(local_selectivity n) {
      n.f0.accept(this);
      curLCSelectivity = new Double(n.f0.toString()).doubleValue();
   }

   /**
    * f0 -> <INTEGER>
    *       | <FLOAT>
    *       | <SQLSTRING>
    */
   public void visit(value n) {
      n.f0.accept(this);
      curValue = ((NodeChoice)n.f0).choice.toString();
   }
  /**
   * return MKB object.
   */
  public MKB getMKB(){
    return mkb;
  }
}
