////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MW;

import java.io.*;
import java.util.*;
import java.sql.*;
import DSRG.MKB.*;

/**
*
* The lowest layer of the DB connection, right above the ODBC driver. <BR>
* All methods go directly to the DB, every time they are invoked. <BR>
* This is subclassed to provide functionality for different DB systems
* in the ISs.
*
* @author Andreas Koeller, November 1997
*/

public abstract class DBManager {

    static final boolean DEBUG = false;

    /**
     *  DB URL
     */
    public String url;
    /**
     *  Username
     */
    public String user;  // would be safe to initialize to eve
    /**
     *  Password
     */
    public String passwd;
    /**
     * The Class name for the appropriate DB driver
     */
    public String className;
    /**
     *  the DB driver
     */
    public Class driver;
    /**
     * the DB connection
     */
    public java.sql.Connection connection = null;

    /**
     * the name of the IS for this DBManager
     */
    public String ISName;

    /**
     * am I connected?
     */
    protected boolean connected = false;
    /**
     * constant for the string Oracle expects for the numeric datatype
     */
    public static final String numeric_datatype = "NUMBER";
    /**
     * constant for the string Oracle expects for the date datatype
     */
    public static final String date_datatype = "DATE";
    /**
     * constructior.<BR>
     * Initializes the "connected" variable.
     */
    public DBManager(){
        connected = false;
    }
  
    /**
     * am I connected?
     */
    public boolean isConnected(){
        return connected;
    }

    /** returns the ISName for this DBManager
     */
    public String getISName() {
        return ISName;
    }
  
    /**
     * gets Relation names for this IS. <BR>
     *
     * 10/6/97 koeller
     */
  
  
    public Vector getRelationNames () {
    
        java.sql.PreparedStatement stmt = null;
        java.sql.ResultSet rs = null;
        java.sql.ResultSetMetaData rsmt = null;
    
        Vector result = new Vector();
    
        try {
            stmt =
                connection.prepareStatement ("select tname " +
                                             "from tab");
      
            rs = stmt.executeQuery ();
            rsmt = rs.getMetaData ();
      
            while (rs.next ()) {
                result.addElement(rs.getString (1));
            }
      
            rs.close ();
            stmt.close ();
        }
        catch (java.sql.SQLException exc) {
        }
    
        return result;
    }
  
    /**
     *
     * returns a vector of strings representing attribute names in a certain relation<BR>
     *
     *
     * 10/6/97 koeller
     */
  
  
    public Vector getAttributeNames (String RelationName) {
    
        java.sql.PreparedStatement stmt = null;
        java.sql.ResultSet rs = null;
        java.sql.ResultSetMetaData rsmt = null;
    
        Vector result = new Vector();
    
        try {
            stmt =
                connection.prepareStatement ("select column_name from cols where table_name " +
                                             "= ?");
      
            stmt.setString(1, RelationName);
            rs = stmt.executeQuery ();
            rsmt = rs.getMetaData ();
      
            while (rs.next ()) {
                result.addElement(rs.getString (1));
            }
      
            rs.close ();
            stmt.close ();
        }
        catch (java.sql.SQLException exc) {
            System.out.println("SQL Exception in getAttributeNames: " + exc.toString());
        }
    
        return result;
    }
  
  
  
    /**
     * a string representing the type <I>(in the syntax of the underlying DB!)</I>
     * of the attribute <CODE>attributeName</CODE> in the relation <CODE>relationName</CODE><BR>
     * 10/6/97 koeller
     */
  
  
    public String getAttributeType (String relationName, String attributeName) {
    
        java.sql.PreparedStatement stmt = null;
        java.sql.ResultSet rs = null;
        java.sql.ResultSetMetaData rsmt = null;
    
        String result = null;
    
        try {
            stmt =
                connection.prepareStatement (   "select data_type, data_length"
                                                +  " from cols where "
                                                +  " table_name = ? "
                                                +  " and column_name = ?");
      
            stmt.setString(1, relationName);
            stmt.setString(2, attributeName);
      
            rs = stmt.executeQuery ();
            rsmt = rs.getMetaData ();
      
            if (rs.next ()) {
                result=   rs.getString(1)  // data_type
                    + "("
                    + rs.getString(2)  // data_length
                    + ")" ;
        
                //this becomes something like NUMBER(38) or so,... hopefully
            }
      
            rs.close ();
            stmt.close ();
        }
        catch (java.sql.SQLException exc) {
            System.out.println("SQL Exception in getAttributeType: " + exc.toString());
        }
    
        return result;
    }
  
    /**
     *
     * returns a vector of strings, which represent actual data values.
     * (regardless of their type in the DB there are always strings coming back!)<BR>
     *
     * 10/6/97 koeller
     */
  
  
    public Vector getAttributeExtent (String RelationName, String AttributeName) {
    
        java.sql.PreparedStatement stmt = null;
        java.sql.ResultSet rs = null;
        java.sql.ResultSetMetaData rsmt = null;
    
        Vector result = new Vector();
    
        // no preparedStatement possible, since Oracle doesn't understand identifier names
        // as parameters... same for Access, I guess
    
        try {
            stmt =
                connection.prepareStatement ("select "+AttributeName+" from "+RelationName);
      
            rs = stmt.executeQuery ();
            rsmt = rs.getMetaData ();
      
            while (rs.next ()) {
                result.addElement(rs.getString (1));
            }
      
            rs.close ();
            stmt.close ();
        }
        catch (java.sql.SQLException exc) {
            System.out.println("SQL Exception in getAttributeExtent: " + exc.toString());
        }
    
        return result;
    }


    /** executes a SQL-Query against one database. The attribute names
        in the returning table are composed of relation names and
        attribute names of the attributes in the select-clause. */
    public Table executeQuery(String SQL) {
        return privateExecuteQuery(SQL,true);
    }
  
    /** executes a SQL-View-Query against one database (which should
        be the local database, of course). The attribute names in the
        returning table are only the attribute names of the attributes
        in the select-clause. Otherwise, this method is equivalent to
        @see executeQuery(String SQL)*/
    public Table executeViewQuery(String SQL) {
        return privateExecuteQuery(SQL,false);
    }
  
    /**
     *
     * returns a single Table object not tied to an IS. <BR> expects
     * SQL in the dialect the DB understands, so for now we need to
     * stick with basic SQL (which we do in the theory...).<BR> If the
     * second parameter is true, the Table object contains attribute
     * names that are composed of relation and attribute names of the
     * SQL-Query in the oaramater. Otherwise, the attribute names are
     * only attribute names, and do not contain relation names. This
     * method should only be called by the <code>executeQuery()</code>
     * and <code>executeViewQuery()</code>-methods.
     *
     * @author koeller (04/07/1999) */
    
    
    private Table privateExecuteQuery(String SQL, boolean withRelationNames) {
        
        java.sql.PreparedStatement stmt = null;
        java.sql.ResultSet rs = null;
        java.sql.ResultSetMetaData rsmt = null;
        
        //        Relation relation= new Relation("QUERY",this);
        Table re = new Table();
        String attribute=null;
        Integer attributeSize;
        Vector attributes=new Vector();
        Vector attributeSizes=new Vector();
        String s=null;
        int bugfix_Columnsize;
        
        try {
            if (DEBUG) System.out.println("Debug: DBManager.newExecuteViewQuery: SQL is: " + SQL);
            if(connection == null)
                System.err.println("Debug: DBManager.executeViewQuery connect is null!");
            
            stmt =  connection.prepareStatement (SQL);
            
            rs = stmt.executeQuery ();
            rsmt = rs.getMetaData();

            for (int i=1; i<=rsmt.getColumnCount(); i++) {
                // another bug: for type "date", the driver returns 1
                // as the column size. Format is : "1970-01-31
                // 00:00:00.0" = 21 chars
                
                if (rsmt.getColumnTypeName(i).equals("DATE"))
                    bugfix_Columnsize=21;
                else bugfix_Columnsize = rsmt.getColumnDisplaySize(i);
                if (withRelationNames)
                    attribute = new String(   getTableName(SQL, i) +
                                              // not implemented in the driver
                                              // so we bugfix it below... :-(
                                              "_" +
                                              rsmt.getColumnName(i));
                else
                    attribute = new String(rsmt.getColumnName(i));
                attributeSize = new Integer(bugfix_Columnsize);
                attributes.addElement(attribute); 
                // addElement adds to end of vector - thus preserving order
                attributeSizes.addElement(attributeSize);
            }
            
            re.setAttributeNames(attributes); // set attributes 
            // in Table

            if (DEBUG) System.err.println("DBManager::computeExtent: AttributeNames are: "+attributes);
            re.setAttributeSizes(attributeSizes);

            Vector tmpVector;
            
            while (rs.next()) {
                
                tmpVector = new Vector(rsmt.getColumnCount());
                for (int i=1; i<=rsmt.getColumnCount(); i++) {
                    s=rs.getString(i);
                    //attribute=(Attribute)attributes.elementAt(i-1);
                    tmpVector.addElement(s);
                }
                re.addTuple(new Tuple(re, tmpVector));
                
            }
            //System.out.println("Debug: DBManager.executeViewQuery step4.");
            
            if (DEBUG) System.out.println("NEW EXTENT:\n"+re.toString());
            rs.close ();
            stmt.close ();
        }
        catch (java.sql.SQLException exc) {
            System.out.println("SQL Exception in executeViewQuery: " + exc.toString());
        }catch (Exception e){
            System.out.println("Debug:DBManager.executeViewQuery: Error:" + e.getMessage());
            e.printStackTrace();
        }
        
        return re;
    }
    
    /** send an update :)*/  
    public boolean sendUpdate(String relationName, int count, String update) {
        java.sql.PreparedStatement stmt = null;
        int result=0;
        
        
        try {
            if(EVEDemo.MW.Main.DEBUG)
                System.out.println("Debug: DBManager.sendUpdate to Relation: " + relationName);
            //System.out.println("Debug: DBManager.executeViewQuery step1.");
            
            if(connection == null)
                System.out.println("Debug: DBManager.executeViewQuery connect is null!");
            
            // prepare insert statement. simply replace "|" with commas.
      
            String statement;
      
            // get the schema of this relation specified by the relationName.
            stmt = connection.prepareStatement ("select * from " + relationName);
            ResultSet rs = stmt.executeQuery();
            ResultSetMetaData rsmd = rs.getMetaData();
            int colcnt = rsmd.getColumnCount();
            StringTokenizer stok = new StringTokenizer(update, "|");
      
            StringBuffer where = new StringBuffer();
            for(int i = 1; i <= colcnt; i++){
                if(stok.hasMoreTokens())
                    where.append(rsmd.getColumnName(i) + " = " + stok.nextToken());
                if(i != colcnt )
                    where.append(" and ");
            }
      
            if (count<0) {
                statement = "delete from " +
                    relationName + 
                    " where " + where;
                if(EVEDemo.MW.Main.DEBUG)
                    System.out.println("DEBUG: DBManager:sendUpdate() statement is:" + statement);
            }else{
                statement = "insert into " +
                    relationName +
                    " values(" +
                    update.replace('|',',') +
                    ")";
            }
            // close the statement to save the cursor
            stmt.close();
            //System.out.println("Sending update to DB: "+statement);
      
            stmt =
                connection.prepareStatement ( statement	);
      
            // execute update "count" times
            count = count < 0? -count:count;
            for (int i=0; i < count; i++) {
                result = stmt.executeUpdate ();
                if (result == 0) break;
            }

            stmt.close();
      
        }
        catch (java.sql.SQLException e) {
            System.out.println(e.toString());
        }
    
        // if all were successful - return true, otherwise false
        //System.out.println("Result of update: "+result);
        return (result != 0);    
    
    }


    /** interface for the schema change operations */
  
    public abstract boolean addRelation (String relationName, Vector attName, Vector attTypes);
    public abstract boolean deleteRelation (String name, boolean silent) ;
    public abstract boolean renameAttribute (String RelationName, String oldName, String newName);
    public abstract boolean addAttribute (String RelationName, String AttributeName,String type, int size);
    public abstract boolean deleteAttribute (String RelationName, String attributeName);
    public abstract boolean renameRelation (String oldName, String newName);
    //  public abstract boolean storeRelation(Relation relation);
    public abstract boolean storeRelation(Table table);
  


    //============================================================================
    //================                private functions                  =================
    //====================================================================================
    //====================================================================================
    //====================================================================================
    //====================================================================================
    //====================================================================================
  
  
    /**
     * Helper function. <BR>
     * this strips one attribute from a schema obtained by getSchema(),
     * kind of awkward...
     *
     */
  
    public String stripAttributeFromSchema(String schema, String attributeName) {
        // some String-acrobatics..., oops, we need a non-case-sensitive version
    
        String upc_schema = new String(schema.toUpperCase());
        String upc_attributeName = new String(attributeName.toUpperCase());
    
    
        int position = upc_schema.indexOf(upc_attributeName);
    
        if (position < 0) return null;
    
        int end = upc_schema.indexOf(',', position);
    
        if (end==-1) // last attribute in schema
      
            return upc_schema.substring(0,position-1)+")";
        else return (  upc_schema.substring(0,position-1)
                       + upc_schema.substring(upc_schema.indexOf(',', position)));
    
    
    }
  
  
    /**
     * determines which table a certain attribute comes from. <BR>
     * looks like this function (ResultSetMetaData.getTableName(int i))
     * is not implemented in the JDBC driver, so I have to parse the
     * table name from the SQL-String. Great!!!! <br>
     * it's only partly implemented - it needs: <br>
     * <code>{&lt;table_name&gt;.&lt;attribute_name&gt;,}&lt;table_name&gt;.&lt;attribute_name&gt;</code>
     * as a format.
     * <BR>
     * It also doesn't work when there is a "*" in the select clause
     * 6/3/98 koeller
     * 
     * I'm going to have to fix the *-Problem, but it's only working with one relation then
     * 9/24/98 koeller
     *
     */
  
    private String getTableName(String SQL, int index) {
        // index = 1...getColumnCount()
        int pos = 0;
    
        if (SQL.indexOf("*")!=-1) 
            return (SQL.substring(SQL.toUpperCase().indexOf("FROM")+1));
    
        if (index==1)
            return (SQL.substring(6,SQL.indexOf(".")).trim());  // after "select"
        else {
            // the string after the next comma after the (index-1)st occurrence of "."
            for (int i=1; i<=index-1; i++)
                pos = SQL.indexOf(".",pos+1);
            return (SQL.substring(SQL.indexOf(",",pos)+1,SQL.indexOf(".",pos+1)).trim());
        }
    }
  
  
    /**
     * Helper function.  <BR>
     * this renames one attribute in a schema obtained by getSchema(),
     *
     *
     */
  
    public String renameAttributeInSchema(String schema, String oldName, String newName) {
        // some more String-acrobatics...
    
        int position = schema.toUpperCase().indexOf(oldName.toUpperCase().trim());
    
        if (position < 0) return null;
    
        int end = schema.indexOf(' ', position);
        // just the name, the type stays the same !!!!!
    
        if (end==-1) // error, can't happen, I thought, until it did
      
            return null;
    
        else return (  schema.substring(0,position)
                       + newName.trim()
                       + schema.substring(end) );
    
    
    }
  
  
  
  
    /**
     * Helper function. <BR> This turns a schema obtained by getSchema
     * into a list of attribute names without the types, to include
     * them in a select clause when copying tables
     * */
  
    String makeSelectClause(String schema) {
    
        String result = new String();
    
        int position=0;
        int position2=0;
    
        for (position=0; schema.charAt(position)<'A'; position++); 
        // find first real character
    
        while (position!=0) { // while there are still commas in the string
            
            // it works only since there is at least one '(' in the
            // beginning in the string !!!
      
      
            position2 = schema.indexOf(' ',position+1); // end of identifier
      
            if (position2==-1) // only one attribute, special case
                position2=schema.length()-1;
      
            result =   result
                + schema.substring(position, position2).trim()
                + " ,";
      
            position = schema.indexOf(',', position2)+1; // start of next identifier
        }
    
        return result.substring(0,result.length()-2);
    
    }
  
}
