////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.VKB;

import java.util.*;
import java.io.*;

/**
 *
 * this class is needed to support the QC-Model. It contains the basic
 * values to compute a QC-Value and the function to actually compute
 * it. Currently the base values are hardcoded, that is soon to
 * change.
 * @author koeller, 030799 */

public class QC {
    
    /** 
     * the degrees of divergence */
    private double DD_attr, DD_ext;

    /**
     * the trade-off parameters for the degrees of divergence */
    private double rho_attr, rho_ext;
    
    /**
     * the values for the cost dimensions */
    private long CF_T, CF_M, CF_IO;
    
    /**
     * the tradeoff parameters for the cost dimensions */
    private double cost_T, cost_M, cost_IO;
    
    /**
     * the tradeoff parameters for the two dimensions */
    private double rho_quality, rho_cost;
    
    /**
     * for normalization of the cost, I'll keep a maximum cost that is
     * used in the computation of the complete value and is global for
     * all QC-instances */
    private static double maxCost = -1.0;

    // Comment: rho_ext and rho_cost shouldn't exist but be computed as
    // 1-rho_<other>. But I don't want to prefer one over the other.
    
    /**
     * This is the testing constructor. Values are hardcoded and
     * random. The total QC-Value always adds up to something between
     * zero and one.  */
    
    public QC(/*not sure yet which parameters it needs*/) {
	// Hardcoding some values here for development purposes
	rho_attr = 0.5; 
	rho_quality = 0.9;
	computeOtherValues();
	setRandomValues();
    }

    /**
     * This constructor takes the preset values for the tradeoff
     * factors. rho_cost and rho_ext are computed from rho_quality and
     * rho_attr, respectively.  */
    
    public QC(double my_rho_attr, double my_rho_quality) {
	// Hardcoding some values here for development purposes
	rho_attr = my_rho_attr;
	rho_quality = my_rho_quality;
	computeOtherValues();
	setRandomValues(); // remove this later!!!
    }

    /**
     * This constructor takes the preset values for the tradeoff
     * factors including the cost tradeoff factors ("unit
     * costs"). rho_cost and rho_ext are computed from rho_quality and
     * rho_attr, respectively.<BR>  This is the final constructor, no
     * random values are assigned here (everything comes from the
     * Version class) */
    
    public QC(double my_rho_attr, double my_rho_quality,
              double my_cost_M, double my_cost_T, double my_cost_IO) {
	rho_attr = my_rho_attr;
	rho_quality = my_rho_quality;
	computeOtherValues();
        cost_M  = my_cost_M;
        cost_T  = my_cost_T;
        cost_IO = my_cost_IO;
    }


    /** 
     * This method computes rho_cost and rho_ext from the other two
     * tradeoff values */
    private void computeOtherValues() {
	rho_cost = 1.0 - rho_quality;
	rho_ext = 1.0 - rho_attr;
    }
    
    /** 
     * private method during testing to set some random QC-Values */

    private void setRandomValues() {
	DD_attr = java.lang.Math.random();
	DD_ext  = java.lang.Math.random();
	CF_T = 817265;
	CF_M = 98247;
	CF_IO = 891724687;
	// make sure costs are not larger than one
	cost_T = java.lang.Math.random()/3.0  / CF_T;
	cost_M = java.lang.Math.random()/3.0  / CF_M;
	cost_IO = java.lang.Math.random()/3.0 / CF_IO;
    }

    /**
     * compute the total QC-Value from the base values. simple
     * mult-and-add. Test QC-Value (tradeoff parameters, base values)
     * for plausibility.
     * @return QC-Value. -1 if QC-Value is not valid. */
    
    public double getQCValue () {
	if (!isQCValid()) 
	    return -1.0;
	else
	    return  ( 1 - 
                      ( rho_quality * getQuality()
                        +
                        rho_cost * getCost()
                      ) 
                    );
    }

    // Standard get-and-set methods

    /**
     * returns the Degree of Divergence of the Attributes */
    public double getDDattr () {
	return DD_attr;
    }


    /**
     * returns the Degree of Divergence of the Extent */
    public double getDDext () {
	return DD_ext;
    }
    
    /** sets the two Degrees of Divergence at once
     * @return QC-Validity check bit */
    public boolean setDD( double myDD_attr, double myDD_ext) {
        DD_attr = myDD_attr;
        DD_ext = myDD_ext;
        return isQCValid();
    };

    /**
     * returns the tradeoff-parameter for the Degree of Divergence of
     * the Attributes */
    public double getRhoAttr () {
	return rho_attr;
    }


    /**
     * returns the tradeoff-parameter for the Degree of Divergence of
     * the Extent */
    public double getRhoExt () {
	return rho_ext;
    }


    /**
     * returns the Number of Messages */
    public long getCFM () {
	return CF_M;
    }


    /**
     * returns the Number of Bytes Transferred */
    public long getCFT () {
	return CF_T;
    }


    /**
     * returns the Number of I/O-Operations */
    public long getCFIO () {
	return CF_IO;
    }

    /** sets the three cost factors at once.
     * @return QC-Validity check bit */
    public boolean setCF(long myCF_M, long myCF_T, long myCF_IO) {
        CF_M  = myCF_M;
        CF_T  = myCF_T;
        CF_IO = myCF_IO;
        double rawCost = 
            (double) CF_M * cost_M + 
            (double) CF_T * cost_T + 
            (double) CF_IO * cost_IO;
        if (maxCost<rawCost)
            maxCost = rawCost;
        return isQCValid();
    }

    /**
     * returns the tradeoff parameter ("unit cost") for the Number of
     * Messages */
    public double getCostM () {
	return cost_M;
    }

    /**
     * returns the tradeoff parameter ("unit cost") for the Number of
     * Bytes Transferred */
    public double getCostT () {
	return cost_T;
    }

    /**
     * returns the tradeoff parameter ("unit cost") for the Number of
     * I/O-Operations */
    public double getCostIO () {
	return cost_IO;
    }


    /**
     * returns the tradeoff-parameter for the total Degree of
     * Divergence (Quality) */
    public double getRhoQuality () {
	return rho_quality;
    }


    /**
     * returns the tradeoff-parameter for the total Cost */
    public double getRhoCost () {
	return rho_cost;
    }


    /** 
     * gets the total cost by multiplying the costs with their cost
     * factors and adding them. normalizes with maxCost */
    
    public double getCost() {
        if (maxCost < 0)
            return (double) CF_M * cost_M + 
                (double) CF_T * cost_T + 
                (double) CF_IO * cost_IO;
        else
            return ((double) CF_M * cost_M + 
                    (double) CF_T * cost_T + 
                    (double) CF_IO * cost_IO)/maxCost;
    }


    /** 
     * gets the total quality by multiplying the degrees of divergence
     * with their tradeoff factors and adding them */
    
    public double getQuality() {
	return DD_attr * rho_attr + 
	       DD_ext * rho_ext;
    }



    /**
     * checks if the tradeoff parameters are valid (i.e., the pairs
     * add up to one) */

    protected boolean isQCValid() {
	
	return ( rho_attr >= 0.0
		 && rho_ext  >= 0.0
		 // fuzzy float-compare
		 && java.lang.Math.abs ( rho_ext + rho_attr - 1.0 ) < 1.0E-10 
		 // careful, this is not in the TR: Cost is normalized.
		 && getCost() >= 0.0 // testing for negativity
		 && getCost() * rho_cost <= 1.0 // is it "normalized"?
		 && rho_quality >= 0.0
		 && rho_cost >= 0.0
		 && java.lang.Math.abs ( rho_quality + rho_cost -1.0 ) < 1.0E-10 
		 ) ;
    }

    /** returns a string representation of the whole QC-object for
     *  testing purposes. The format is <CODE> QC = 1- (0.3 * ( 0.4 *
     *  0.85 + 0.6 * 0.9 ) + 0.7 * (0.0001 * 8000 + 0.0002 * 4000 +
     *  0.0004 * 2500)) </CODE> or <CODE> QC = 1- (rho_quality * (
     *  rho_attr * DD_attr + rho_ext * DD_ext ) + rho_cost * (cost_M *
     *  CF_M + cost_T * CF_T + cost_IO * CF_IO))</CODE> */

    public String toString() { 
        return "QC = 1- (" + rho_quality + "* (" + rho_attr + "*" + DD_attr + 
            " + " + rho_ext + " * " + DD_ext + ") + " + rho_cost + " * (" +
            cost_M + " * " + CF_M + " + " + cost_T + " * " + CF_T + " + " 
            + cost_IO + " * " + CF_IO + "))"; 
    }

    /** resets the max.cost. To be called before the creation of a new
        set of QC-Values */
    public static void resetMaxCost() {
        maxCost = -1.0;
    }
}

