////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.VKB;

import java.util.*;
import java.io.*;
import DSRG.MKB.MKB;

/**
 *
 * this class is needed to support the QC-Model. It contains a view
 * definition, and a vector of suggestions for its rewriting after a
 * certain schema change. Those suggestions are not materialized, but
 * they have "QC-Values", which are objects of another new class,
 * called "QC".
 * @author koeller, 030799 */

public class Version {

    private static final boolean DEBUG = false;
    
    private View_Definition oldDefinition, definition;
    private QC qcValues;
    

    /**
     * the tradeoff parameters for replaceable and non-replaceable
     * attributes, this will not be in the QC-class. not sure if
     * that's good (ask Xin and Elke) */
    private double w_repl, w_nonrepl;

    private MKB myMKB;

    /**
     * the constructor creates a version of a view.
     * @author koeller */

    public Version(View_Definition myOldDefinition, 
                   View_Definition myDefinition, 
                   QC myQCValues, MKB theMKB) {
        myMKB = theMKB;
        w_repl = 0.5;
        w_nonrepl = 1 - w_repl;
        oldDefinition = myOldDefinition;
	definition = myDefinition;
	qcValues = myQCValues;
        qcValues.setDD(computeDDattr(), computeDDext());
        qcValues.setCF(computeCF_M(), computeCF_T(), computeCF_IO());
    }

    /** compute DDattr from the View Definition
     * @return DDattr-value */
    private double computeDDattr() {
        double QVold = computeQV(oldDefinition);
        if (QVold == 0) 
            return 0;
        else 
            return (QVold - computeQV(definition)) / QVold;
    }
    
    /** compute Q(V) --- the intermediate value needed to compute
     *  DD_attr */

    private double computeQV(View_Definition vd) {
        int number_replaceable = 0, number_nonreplaceable = 0;
        Vector selectClause = vd.getClause(DSRG.VKB.View_Definition.SELECT);
        Attribute_Component tmp = null;
        Enumeration e = selectClause.elements();
        while (e.hasMoreElements()) {
            tmp = (Attribute_Component)e.nextElement();
            if (tmp.getDispensable()) {
                if (tmp.getReplaceable())
                    number_replaceable++;
                else
                    number_nonreplaceable++;
            }
        }
        return number_replaceable * w_repl +
            number_nonreplaceable * w_nonrepl;
    }

    /** compute DDext from the ViewDefinitions. For POC without
     *  approximate PC-constraints, it is sufficient to count the
     *  smaller of the two views as overlap. That's all we need. So
     *  we'll just get a size of the view before and after view
     *  synchronization. For the latter, we use the same algo as for
     *  CF_T (for current_no_of_tuples) :) */
    private double computeDDext() {
        
        // now, no_of_tuples contains the size of the new view (hopefully)
        long oldSize = viewSizeEstimate(oldDefinition);
        long newSize = viewSizeEstimate(definition);


        if (DEBUG) System.out.println("QC: DDExt: "+oldSize+" / "+newSize);
        
        double DDext;
        if (definition.getVExtent()==View_Definition.SUBSET)
            {
                if (oldSize  == 0) DDext = 0.0;
                else
                    DDext = (double)(oldSize - newSize) / (double)oldSize;
            }
        else
            {
                if (newSize == 0 ) DDext = 0.0;
                else 
                    DDext = (double)(newSize - oldSize) / (double)newSize;
            }
        if (DEBUG) System.out.println("DDext: "+DDext);
        if (DDext<0.0) DDext = 0.0;
        if (DDext>1.0) DDext = 1.0;
        
        return DDext;

    }

    /** private method to compute the number of tuples in a view (by
        multiplying selectivities and relation sizes */

    private long viewSizeEstimate(View_Definition def) {
        
        // this is the size of the current query including the update
        // tuple (No. of tuples, not bytes !!!!)
        long current_no_of_tuples=1;

        // temporary variables
        double sigma;
        double J;


        // outer loop: InformationSources in the query
        //
        Enumeration ISs = 
            ((Vector) def.getISs()).elements();
        // trying to support local conditions.. Get all Conditions in
        // this VD's where-clause
        Vector allConditions = def.getClause(DSRG.VKB.View_Definition.WHERE);

        String ISName = null;
        String prevISName = null;
        while (ISs.hasMoreElements()) {
            prevISName = ISName;
            ISName = (String)ISs.nextElement();
            
            
            
            // reset for each IS
            sigma = 1;
            J = 1;
            
            // then compute right summand

            // first, get all relations
            Enumeration relations = 
                ((Vector) def.getClause
                 (DSRG.VKB.View_Definition.FROM)).elements();
            Relation_Component tempRelC;
            Vector relations_in_IS=new Vector();
            while (relations.hasMoreElements()){
                // if relation exists in this IS
                tempRelC = (Relation_Component) relations.nextElement();
                if (ISName.equals(tempRelC.getISName()))
                    relations_in_IS.addElement(tempRelC);
            }
            // now we have a vector of USED relations in this IS
            //  loop through them and compute sigma and J
            String relName,prevRelName=null;
            
            for (int i = 0 ; i < relations_in_IS.size(); i++) {
                relName = ((Relation_Component)relations_in_IS
                           .elementAt(i)).getName();

                // for the current ISName and Relationname, we search
                // for local Conditions....
                Condition c; double selectivity;
                for (int i1 = 0 ; i1 < allConditions.size(); i1++ ) {
                    c = (Condition)((Condition_Component)allConditions
                                    .elementAt(i1)).getCondition();
                    if (c instanceof Local_Condition) {
                        // these are all from where-clauses. So if an
                        // attribute is in VD, it's also in this VD's
                        // WHERE-clause
                        
                        DSRG.MKB.Attribute a = c.getLeft();
                        if (def.contains(a) && a.getRelation().getName().equals(relName)) {
                            selectivity = 
                                myMKB.
                                getLocalConditionSelectivity(
                                  a.getRelation().getISName(), 
                                  a.getRelation().getName(),
                                  a.getName(),
                                  ((Local_Condition)c).getRight().toString()
                                  );
                            // and take their selectivities into account
                            sigma *= selectivity;
                        }
                    }
                }
                
        
                
                if (DEBUG) System.out.println("QC.DDext: J="+J);
                if (DEBUG) System.out.println("QC.DDext: sigma="+sigma);
                J *= myMKB.getRelationSize(ISName, relName);
                if (i>=1)
                    J *= myMKB.getJoinSelectivity(ISName, prevRelName, 
                                                  ISName, relName);
                prevRelName = relName;
                // and cumulate attribute size for each relation. We ignore
                // doubling (join) columns
            } // for-loop
            
            // now this is sigma_is1....sigma_isn * J_is1...J_isn
            current_no_of_tuples *= sigma * J;
        }


        // bugfix #3869827 We need bridges between the ISs, so we
        // take the inter-IS-JSs into account
        Condition c; double selectivity;
        for (int i1 = 0 ; i1 < allConditions.size(); i1++ ) {
            c = (Condition)((Condition_Component)allConditions
                            .elementAt(i1)).getCondition();
            if (c instanceof Join_Condition) {
                // these are all from where-clauses.
                
                DSRG.MKB.Attribute a = c.getLeft();
                DSRG.MKB.Attribute b = ((Join_Condition)c).getRight();
                DSRG.MKB.Relation ra = a.getRelation();
                DSRG.MKB.Relation rb = b.getRelation();
                DSRG.MKB.IS isa = ra.getIS();
                DSRG.MKB.IS isb = rb.getIS();
                
                if (isa != isb) {
                    double interJS= myMKB.getJoinSelectivity(isa.getName(),
                                                             ra.getName(),
                                                             isb.getName(),
                                                             rb.getName());
                    
                    current_no_of_tuples *= interJS;
                }
            }
        }
            
        
        if (DEBUG) {
            System.out.println("QC (DDext) View:"+def.getName());
            System.out.println("Current Number of Tuples :"
                               + current_no_of_tuples);
        }
        
        
        return current_no_of_tuples;
    }

    /** compute number of messages. Simplified over the paper (the
     *  case of no other relations in the updated IS is not covered,
     *  i.e., we always have twice as many messages as information
     *  sources. The error should be minimal. Maybe I'll fix that
     *  later.*/
    
    private long computeCF_M () {
        return 2 * definition.getISs().size();
    };


    /** Compute number of transferred bytes.<BR> We need to compute
     * how many bytes are transferred between view site and ISs. So we
     * take an update in relation R1, send it to the site of Relation
     * R2, get the result, and so on. We have to somehow take into
     * account that several relations can reside in the same IS. <BR>
     * As a simplification, I won't take into account where the update
     * came from. I will just assume the cost for a complete
     * recomputation of one tuple of the leftmost relation in the
     * view. Will be extended later.*/

    private long computeCF_T () {
        // ok, all we have is the old and new view definition.
        // we then compute the average for updates in all relations
        // involved, that way we should get a useful estimate of CF_T.
        // so where do we get the join selectivities and relation
        // sizes? :)
        
        // this is a property of the new view definition.

        // this is the total sum of all bytes
        long sum_bytes = 0;
        
        // this is the size of the current query including the update
        // tuple (No. of tuples, not bytes !!!!)
        long current_no_of_tuples=1;

        //this is the width of the current query (combined attribute size)
        int attr_size=0; // should be the size of the update tuple,
                         // currently ignored, since we don't consider
                         // the updated-relation separately :)

        // temporary variables
        double sigma;
        double J;


        // outer loop: InformationSources in the query
        //
        Enumeration ISs = 
            ((Vector) definition.getISs()).elements();
        
        Vector allConditions = definition.getClause(DSRG.VKB.View_Definition.WHERE);

        String ISName = null;
        String prevISName = null;
        while (ISs.hasMoreElements()) {
            prevISName = ISName;
            ISName = (String)ISs.nextElement();
            if (prevISName != null) {

                // bugfix #3869827 We need bridges between the ISs, so we
                // take the inter-IS-JSs into account
                Condition c; double selectivity;
                for (int i1 = 0 ; i1 < allConditions.size(); i1++ ) {
                    c = (Condition)((Condition_Component)allConditions
                                    .elementAt(i1)).getCondition();
                    if (c instanceof Join_Condition) {
                        // these are all from where-clauses.
                        
                        DSRG.MKB.Attribute a = c.getLeft();
                        DSRG.MKB.Attribute b = ((Join_Condition)c).getRight();
                        DSRG.MKB.Relation ra = a.getRelation();
                        DSRG.MKB.Relation rb = b.getRelation();
                        DSRG.MKB.IS isa = ra.getIS();
                        DSRG.MKB.IS isb = rb.getIS();
                
                        if (isa != isb) {
                            double interJS= myMKB.getJoinSelectivity(isa.getName(),
                                                             ra.getName(),
                                                             isb.getName(),
                                                             rb.getName());
                    
                            current_no_of_tuples *= interJS;
                        }
                    }
                }
            }

            //according to formula: add left summand of this IS's line
            // (same as before)
            sum_bytes += (current_no_of_tuples*attr_size);
            // reset for each IS
            sigma = 1;
            J = 1;
            
            // then compute right summand

            // first, get all relations
            Enumeration relations = 
                ((Vector) definition.getClause
                 (DSRG.VKB.View_Definition.FROM)).elements();
            Relation_Component tempRelC;
            Vector relations_in_IS=new Vector();
            while (relations.hasMoreElements()){
                // if relation exists in this IS
                tempRelC = (Relation_Component) relations.nextElement();
                if (ISName.equals(tempRelC.getISName()))
                    relations_in_IS.addElement(tempRelC);
            }
            // now we have a vector of USED relations in this IS
            //  loop through them and compute sigma and J
            String relName,prevRelName=null;
            for (int i = 0 ; i < relations_in_IS.size(); i++) {
                relName = ((Relation_Component)relations_in_IS.elementAt(i)).getName();

                // for the current ISName and Relationname, we search
                // for local Conditions....
                Condition c; double selectivity;
                for (int i1 = 0 ; i1 < allConditions.size(); i1++ ) {
                    c = (Condition)((Condition_Component)allConditions.elementAt(i1)).getCondition();
                    if (c instanceof Local_Condition) {
                        // these are all from where-clauses. So if an
                        // attribute is in VD, it's also in this VD's
                        // WHERE-clause
                        
                        DSRG.MKB.Attribute a = c.getLeft();
                        if (definition.contains(a)  && a.getRelation().getName().equals(relName)) {
                            selectivity = 
                                myMKB.
                                getLocalConditionSelectivity(
                                  a.getRelation().getISName(), 
                                  a.getRelation().getName(),
                                  a.getName(),
                                  ((Local_Condition)c).getRight().toString()
                                  );
                            // and take their selectivities into account
                            sigma *= selectivity;
                        }
                    }
                }
                
        
                
                J *= myMKB.getRelationSize(ISName, relName);
                if (i>=1)
                    J *= myMKB.getJoinSelectivity(ISName, prevRelName, ISName, relName);
                prevRelName = relName;
                // and cumulate attribute size for each relation. We ignore
                // doubling (join) columns
                attr_size += myMKB.getAttributeSizeForRelation(ISName, relName);
                if (DEBUG) System.out.println("Attr_size: "+attr_size);
                if (DEBUG) System.out.println("sigma: "+sigma);
                if (DEBUG) System.out.println("J: "+J);

            } // for-loop

            // now this is sigma_is1....sigma_isn * J_is1...J_isn
            current_no_of_tuples *= sigma * J;


            // attr_size should now be correct, sort of :)

            // and add to total
            sum_bytes += ((long)current_no_of_tuples*attr_size);

            if (DEBUG) {
                System.out.println("QC (CF_T):");
                System.out.println("current_no_of_tuples :"+current_no_of_tuples);
                System.out.println("attr_size :"+attr_size);
                System.out.println("sum_bytes :"+sum_bytes);
            }
        }
        
        return sum_bytes;

        // the VM algorithm will always transport whole relations,
        // minus the duplicate attribute. That way, we will never
        // miss an attribute that may be needed for a join later.
    };

    /** compute number of I/O-operations. As a first hack, this is the
     *  number of total bytes divided by the blocking factor :) */
    private long computeCF_IO () {
        return (long)(computeCF_T() / 2048)+1;
    };
    

    /**
     * returns the View_Definition object of this version
     */
    public View_Definition getDefinitionAsObject() {
        return definition;
    }

    /**
     * returns the View_Definition object of the old VD of this
     * version */
    public View_Definition getOldDefinitionAsObject() {
        return oldDefinition;
    }

    /**
     * returns the (SQL-)definition of this version
     * @author koeller
     */
    
    public String getDefinition () {
	return definition.getSQL();
    }
    
    /** returns the QC-Object of this version
     * @author koeller */
    
    public QC getQC() {
	return qcValues;
    }
}
