////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.VKB;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.io.StringBufferInputStream;

import DSRG.VKB.E_SQL.E_SQL;
import DSRG.VKB.E_SQL.SimpleNode;
import DSRG.VKB.E_SQL.ParseException;
import DSRG.VKB.Join_Condition;
import DSRG.VKB.Local_Condition;
import DSRG.VKB.View_Component;
import DSRG.VKB.Attribute_Component;
import DSRG.VKB.Relation_Component;
import DSRG.VKB.Condition_Component;

import DSRG.MKB.MKB;
import DSRG.MKB.Attribute;
import DSRG.MKB.Relation;
import DSRG.MKB.IS;

import DSRG.VSA.Replacement;
/**
* Represents the View definition.
*
* <p>
* It has a small parser in it, so, it can generate from a string of ESQL.
* It contains three lists: select, from and where.
*
* <p> ViewDefinition Overall Function
*   <ul>
*     <li> Parse E-SQL </li>
*     <li> Parse SQL </li>
*     <li> Generate Strongest E-SQL ref: TR-98-4 P20~P23</li>
*     <li> Generate SQL </li>
*     <li> Get Clause of SELECT, FROM, and WHERE </li>
*   </ul>
*/
public class View_Definition implements java.io.Serializable, DSRG.VSA.VSA_View_Definition, Cloneable {

private static final boolean DEBUG = false;

  // Used for QC model
  /**
  * A Vector of Versions of this View-Definition. To be used for
  * selecting a version out of some suggestions after a view
  * synchronization.  */
  Vector versions;

  // Clause type
  /**
  * SELECT here has two meaning. If it used in getClause() function, it
  * means the select clause of view definition. If it used in getPSJList() function,
  * it means the local_conditions in the where clause of view definition.
  */
  public static final int SELECT =0;
  /**
  * FROM clause constant
  */
  public static final int FROM = 1;
  /**
  * WHERE clause constant
  */
  public static final int WHERE = 2;

  // PSJ list type
  /**
  * PROJECT list constant
  */
  public static final int PROJECT = 3;
  /**
  * JOIN list constant
  */
  public static final int JOIN = 4;

  // SET Relationships for View Extent Parameter
  /**
  * VE is SUBSET.
  */
  public static final int SUBSET = 0;
  /**
  * VE is SUPERSET
  */
  public static final int SUPERSET = 1;
  /**
  * VE is EQUIVALENT
  */
  public static final int EQUIVALENT = 2;
  /**
  * VE is DONTCARE
  */
  public static final int DONTCARE = 3;

  /**
  * Name of this ViewDefinition.
  */
  String name;

  /**
  * VE parameter of this View Definition
  */
  int VExtent;

  /**
  * it will keep the order of the select clause. It stores a vector of Attribute_Component
  */
  Vector select;

  /**
  * Vector of FROM clause.
  * The elements are objects of class Relation_Component.
  */
  Vector from;

  /**
  * Vector of Condition clause.
  * It's a list of conditions. Each elememt is an object of Condition_Component.
  */
  Condition_index where;

  /**
  * Set a list of versions of rewritings of this View_Definition by VSA.
  * @param A Vector of Version.
  */
  public void setVersions(Vector vs){
  	versions = vs;
  }
    
    /** return the vector of all versions of this view after view
     * synchronization. This vector is updated from the outside, it's
     * never changed from within this class.
     * @author koeller */
    
  public Vector getVersions() {
	  return versions;
  }
    
  /**
   * Clean the View Definition: 
   * 1. Clean the from clause to remove the useless relations that 
   * only appeared in Where clause. 
   * 2. It also clean the duplicate relations.
   * 3. Clean the Where Clause JC conditions.
   */
  public void clean(){
    Hashtable ht = new Hashtable();

    // 1. Clean duplicate Relation
    for(int i = 0; i < from.size();){
      Relation_Component rc = (Relation_Component)from.elementAt(i);
      Relation r = rc.getRelation();
      if (DEBUG) System.out.println("cleanFromClause: Relation is: " + r);
      // 2.1 if it already exists
      Object obj= ht.get(r);
      if(obj!= null && obj != rc){
      	// remove it
      	from.removeElement(rc);
      	if (DEBUG) System.out.println("cleanFromClause: Relation " + r + " has duplicates, so deleted.");
            }else{
      	// else put it into the hashtable
      	ht.put(r, rc);
      	i++;
      }
    }

    // 2. Clean useless Relation that only appeared once in the whole view.
    for(int i = 0; i < from.size();){
      Relation_Component rc = (Relation_Component)from.elementAt(i);
      Relation r = rc.getRelation();
      if (DEBUG) System.out.println("cleanFromClause: Relation is: " + r);
      Vector v = getComponents(r);
      if(v!=null){
      	if (DEBUG) System.out.println("cleanFromClause: size is: " + v.size());
      	if(v.size() == 1){
      	  from.removeElement(rc);
      	  if (DEBUG) System.out.println("cleanFromClause: Relation " + r + " is cleaned");
      	}else
      	  i++;
      
      }else
      	i++;

    }// endfor

    /*
     * 3. Clean the WHERE clause to remove the useless Join Condition that
     * has same attributes as both side.  */

    // 3.1 Enumerate the where clause
    for(int i = 0; i < where.size();){
      Condition_Component cc = (Condition_Component)where.elementAt(i);
      Condition c = cc.getCondition();
      if (DEBUG) System.out.println("check Condition: " + c);
      // if the condition is Join Condition
      if(c instanceof Join_Condition){
      	// if the left equals the right
      	Join_Condition jc = (Join_Condition)c;
      	if(!jc.valid()){
      	  if (DEBUG) System.out.println("We drop Join Condition: " + jc);
      	  where.removeElement(cc);
      	}else
      	  i++;
      }else
	      i++;
    }//endwhile
  }

  /**
  * Drop Attributes from Select clause of this view definition.
  * @param r the attribute's relation
  * @param di ESQL dispensable parameter
  * @param re ESQL replaceable parameter
  */
  public void dropAttrSelect(Relation r, boolean di, boolean re){
    // Enumerate elements
    Enumeration enum = select.elements();

    // If we found the elements, then we drop it.
    while(enum.hasMoreElements()){
      Attribute_Component ac = (Attribute_Component)enum.nextElement();

      if(ac.getAttribute().getRelation() == r
        && ac.Dispensable == di
        && ac.Replaceable == re){
          select.removeElement(ac);
      }
    }
  }

  /**
  * Drop Conditions from Where clause of this view definition.
  * @param r the attribute's relation
  * @param di ESQL dispensable parameter
  * @param re ESQL replaceable parameter
  */
  public void dropCondWhere(Relation r, boolean di, boolean re){
    // similar to dropAttrWhere
    if(where == null)
      return;

    Enumeration enum = where.elements();

    while(enum.hasMoreElements()){
      Condition_Component cc = (Condition_Component)enum.nextElement();
      Condition cond = cc.getCondition();
      if(cond.contains(r) && cc.Dispensable == di && cc.Replaceable == re){
        if(cond.getLeft().getRelation() == r){
          where.removeElement(cond);
	}
        else if(cond instanceof Join_Condition)
          if(((Join_Condition)cond).getRight().getRelation() == r){
            where.removeElement(cond);
	  }
      }// endif
    }// endwhile
  }

  /**
  * replace Attributes in Select Clause.
  *
  * @param attr The Attribute to be replaced
  * @param reattr The Substitute Attribute
  * @return If replace successful (view evolvable) return true. That means for
  * (f, t), all attributes can be replaced; otherwise, fail.
  */
  public boolean replaceAttrSelect(Attribute attr, Attribute reattr){
    
    int size = select.size();

    for(int i = 0; i < size; i++){
      Attribute_Component ac = (Attribute_Component)select.elementAt(i);

      if(ac.getAttribute() == attr && ac.Replaceable){
        // Here I need to create a new Attribute_Component component in order to avoid
        // affect other view definitions reference to the same Attribute_Components.
        Attribute_Component tmp = (Attribute_Component)ac.clone();
        tmp.setAttribute(reattr);
        select.setElementAt(tmp, i);
      }else if(ac.getAttribute() == attr && ac.Replaceable == false){
        if (DEBUG) System.out.println("View_Definition:replaceAttrSelect: error AC " + ac + " contains( " + attr + " ) cannot be replaced!");
      }
    }

    return true;
  }

  /**
  * replace Attributes in Select clause.
  *
  * @param r the attribute's relation.
  * @param rep Replace ment of this relation with hashtable of substitute attributes.
  * @param di ESQL dispensable parameter
  * @param re ESQL replaceable parameter
  * @return If replace successfull (view evolvable) return true. That means for
  * (f,t), all attributes should be replaced, otherwise, fail.
  */
  public boolean replaceAttrSelect(Relation r, Replacement rep, boolean di, boolean re){
  
    // If we found the elements, then we replace it.
    for(int i = 0 ;i < select.size(); i++){
      Attribute_Component ac = (Attribute_Component)select.elementAt(i);

      if(ac.getAttribute().getRelation() == r
        && ac.Dispensable == di
        && ac.Replaceable == re){

        Attribute attr;
        if(re){// if replaceable (*, t)
          // if there is a replacement
          if((attr = (Attribute)rep.attr_pairs.get(ac.getAttribute()))!= null){
            if (DEBUG) System.out.println("View_Definition:replaceAttrSelect "
              + "Attribute " + ac.getAttribute() + " replaced to " +
              attr);
      	    Attribute a = ac.getAttribute();
            // replace it.
            Attribute_Component tmp = (Attribute_Component)ac.clone();
            tmp.setAttribute(attr);
            select.setElementAt(tmp, i);
          }
          else if(di){// else if dispensable (t,t) but no replacement
            if (DEBUG) System.out.println("View_Definition:replaceAttrSelect "
              + "Attribute " + ac.getAttribute() + " dropped.");
      	    Attribute a = ac.getAttribute();
            // drop it.
            select.removeElement(ac);
            // because we remove one element, so, we need to go back 1 to access the next 
            // element.
            i--;
          }
          else {// (f,t) cannot find replacement
      	    if (DEBUG) System.out.println("View_Definition:replaceAttrSelect Attribute " + ac.getAttribute() + " has no substitution!");
            return false;
      	  }
        } // endif re
        else if(di){// (t,f)
          if (DEBUG) System.out.println("View_Definition:replaceAttrSelect "
            + "Attribute " + ac.getAttribute() + " dropped.");
      	  Attribute a = ac.getAttribute();
          // drop it.
          select.removeElement(ac);
          // because we remove one element, so, we need to go back 1 to access the next 
          // element.
          i--;
        }
        else {// (f,f) cannot evolve
      	  if (DEBUG) System.out.println("View_Definition:replaceAttrSelect Attribute " + ac.getAttribute() + " is non-replaceable.");
          return false;
      	}
      }// endif (r, di, re)
    }// endwhile

    return true;
  }
  /**
  * replace Attributes in the condition in the Where clause.
  * @param r the attribute's relation.
  * @param rep Replace ment of this relation with hashtable of substitute attributes.
  * @param di ESQL dispensable parameter
  * @param re ESQL replaceable parameter
  * @return If replace successfull (view evolvable) return true. That means for
  * (f,t), all attributes should be replaced, otherwise, fail.
  */
  public boolean replaceCondWhere(Relation r, Replacement rep, boolean di, boolean re){

    if(where == null)
      return true;

    int size = where.size();

    // If we found the elements, then we replace it.
    for(int i = 0 ;i < size; i++){
      Condition_Component cc = (Condition_Component)where.elementAt(i);
      Condition cond = cc.getCondition();

      if(cond.contains(r)
        && cc.Dispensable == di
        && cc.Replaceable == re){

        Attribute attrL =null , attrR =null; // left and right replacement
        boolean replaceable = false; // if condition is replaceable.

      	/* First we check the left common part of Local(Join)Condition
      	   Then we check the Join Condition right part later.
      	*/
      
      	if(re){// if replaceable (*, t)
      	  replaceable = ((attrL = (Attribute) rep.attr_pairs.get(cond.getLeft()))!= null);
      	  // if there is a replacement	    
      	  if(replaceable){
      	    System.out.print("View_Definition:replaceCondWhere "
      			     + "Condition " + cond + " replaced to ");
      	    // replace it.
      	    Condition_Component tmp = (Condition_Component)cc.clone();
      	    Condition tmp1 = tmp.getCondition();
      	    tmp1.setAttribute(attrL);
      	    where.setElementAt(tmp, i);
      	    if (DEBUG) System.out.println(tmp);
      	  }// end if (replaceable)
      	  else if(di){// (t, t) but no replacement
      	    if (DEBUG) System.out.println("View_Definition:replaceCondWhere "
      			       + "Condition " + cond + " dropped.");
      	    // drop it.
      	    where.removeElement(cond);
      	  }// end if (di)
      	  else // (f,t) but no replacement
      	    if (cond instanceof Local_Condition)
      	      return false;
       	}
      	else if(di){//(t,f)
      	  if (DEBUG) System.out.println("View_Definition:replaceCondWhere "
      			     + "Condition " + cond + " dropped.");
      	  // drop it.
      	  where.removeElement(cond);
      	}
      	else//(f,f) cannot evolve
      	  if (cond instanceof Local_Condition)
      	    return false;
    	
      	// if that is Local_Condition, that is replaced, and go to next one
      	if(cond instanceof Local_Condition)
      	  continue;
    
      	// Then, that is Join_Condition, we need to check right part
      	Join_Condition jc = (Join_Condition)cond;
      
      	// if right part of Join Condition is not relation R, we go to next one
      	if(jc.getRight().getRelation() != r)
      	  continue;
    
      	// Now, we need to check right part, to see, if we can replace it.
      	if(re){// if replaceable (*, t)
      	  replaceable = ((attrR = (Attribute)rep.attr_pairs.get(jc.getRight()))!= null);
      	  // if there is a replacement	    
      	  if(replaceable){
      	    System.out.print("View_Definition:replaceCondWhere "
      			     + "Condition " + jc + " replaced to ");
      	    // replace it.
      	    Join_Condition tmp = (Join_Condition)jc.clone();
      	    tmp.setRight(attrR);	    
      	    where.setElementAt(tmp, i);
      	    if (DEBUG) System.out.println(tmp);
      	  }// end if (replaceable)
      	  else if(di){// (t, t) but no replacement
      	    if (DEBUG) System.out.println("View_Definition:replaceCondWhere "
      			       + "Condition " + jc + " dropped.");
      	    // drop it.
      	    where.removeElement(jc);
      	  }// end if (di)
      	  else // (f,t) but no replacement
      	      return false;
      	}
      	else if(di){//(t,f)
      	  if (DEBUG) System.out.println("View_Definition:replaceCondWhere "
      			     + "Condition " + jc + " dropped.");
      	  // drop it.
      	  where.removeElement(jc);
      	}
      	else//(f,f) cannot evolve
    	    return false;
    
      }// endif (r, di, re)
    }// endwhile
    return true;
  }
 
  /**
  * Get Attributes in Select Clause.
  * @param r the attribute's relation
  * @param di ESQL dispensable parameter
  * @param re ESQL replaceable parameter
  * @return a Vector of attributes satisfied the specifications.
  */
  public Vector attrSelect(Relation r, boolean di, boolean re){
    Enumeration enum = select.elements();
    Vector result = new Vector();

    while(enum.hasMoreElements()){
      Attribute_Component ac = (Attribute_Component)enum.nextElement();
      if(ac.getAttribute().getRelation() == r
        && ac.Dispensable == di
        && ac.Replaceable == re){
          result.addElement(ac.getAttribute());
      }
    }

    if(result.isEmpty())
      return null;

    return result;
  }

  /**
  * Get Attributes in Where Clause.
  * @param r the attribute's relation
  * @param di ESQL dispensable parameter
  * @param re ESQL replaceable parameter
  * @return a Vector of attributes satisfied the specifications.
  */
  public Vector attrWhere(Relation r, boolean di, boolean re){
    if(where == null)
      return null;
    Vector result = new Vector();
    Enumeration enum = where.elements();

    while(enum.hasMoreElements()){
      Condition_Component cc = (Condition_Component)enum.nextElement();
      Condition cond = cc.getCondition();

      if(cond.contains(r) && cc.Dispensable == di && cc.Replaceable == re){
        if(cond.getLeft().getRelation() == r)
          result.addElement(cond.getLeft());
        if(cond instanceof Join_Condition)
          if(((Join_Condition)cond).getRight().getRelation() == r)
            result.addElement(((Join_Condition)cond).getRight());
      }
    }

    if(result.isEmpty())
      return null;

    return result;
  }

  /**
  * Return string representation of VExtent parameter
  */
  String VExtentoStr(int ext){
    switch(ext){
      case SUBSET:
        return "SUBSET";
      case SUPERSET:
        return "SUPERSET";
      case DONTCARE:
        return "DONTCARE";
      case EQUIVALENT:
        return "EQUIVALENT";
      default:
        return "ERROR";
    }
  }

  /**
  * Set View Definition's Name
  */
  public void setName(String str){
    name = str;
  }

  /**
  * Set VE parameter of this View Definition
  */
  public void setVExtent(int ext){
    switch(ext){
      case DONTCARE:
      case SUBSET:
      case SUPERSET:
      case EQUIVALENT:
        VExtent = ext;
        break;
      default:
        if (DEBUG) System.out.println("View_Defintion:setVExtent wrong extent value:" + ext);
    }
  }

  /**
  * Set Select Clause
  */
  public void setSelect(Vector clause){
    select = clause;
  }

  /**
  * Set From clause
  */
  public void setFrom(Vector clause){
    from = clause;
  }

  /**
  * Set Where clause
  */
  public void setWhere(Condition_index clause){
    where = clause;
  }

  /**
  * Default construction.
  */
  public View_Definition(){
    versions = null;
    name = null;
    from = null;
    select = null;
    where = null;
    VExtent = DONTCARE;
  }

  /**
  * Parse E-SQL/SQL String and create the View definition object.
  * @param ESQL E-SQL/SQL of this view definition.
  */
  public View_Definition(String ESQL, MKB mkb){
    this();

    StringBufferInputStream input = new StringBufferInputStream(ESQL);
    E_SQL parser = new E_SQL(input);

    SimpleNode root = null;

    try{
      // parse the E_SQL
      root = parser.View_definition();
    }catch(ParseException pe){
      pe.printStackTrace();
      if (DEBUG) System.out.println("View_Definition(ESQL): parse error: " + pe.getMessage());
      return;
    }

    // create the object from the parser
    root.createViewDefinition(this, mkb);

  }

  /**
  * Get E-SQL of this view definition.
  * @return E-SQL of this view definition.
  */
  public String getESQL(){
    int size = 0;
    StringBuffer buf = new StringBuffer("CREATE VIEW\n" + name);

    if(VExtent != DONTCARE)
        buf.append(" ( VE = " + VExtentoStr(VExtent) + " )");

    buf.append(" AS\n");

    // generate select clause
    size = select.size();

    if(size > 0)
    {
      buf.append("SELECT\n");
      // Here , keys stores a list of reference of attributes stored in MKB.
      Enumeration enum = select.elements();
      for(int i = 0; i <size-1; i++)
        if(enum.hasMoreElements())
          buf.append(((Attribute_Component)enum.nextElement()) +",\n ");

      buf.append(((Attribute_Component)enum.nextElement()) +"\n");
    }

    // generate from clause
    size = from.size();

    if(size > 0)
    {
      buf.append("FROM\n");
      Enumeration enum = from.elements();
      for(int i = 0; i <size-1; i++)
        if(enum.hasMoreElements())
          buf.append(((Relation_Component)enum.nextElement()) +",\n ");

      buf.append(((Relation_Component)enum.nextElement()) +"\n");
    }

    if(where != null){
      // generate where clause
      size = where.size();

      if(size > 0)
      {
        buf.append("WHERE\n");
        Enumeration enum = where.elements();
        for(int i = 0; i <size-1; i++)
          if(enum.hasMoreElements())
            buf.append(((Condition_Component)enum.nextElement()) +"\n  AND ");

        buf.append(((Condition_Component)enum.nextElement()) +"\n");
      }
    }

    return buf.toString();
  }

  /**
  * Get SQL of this view definition without the IS information.
  * @return SQL of this view definition.
  */
  public String getSQL(){
    int size = 0;
    StringBuffer buf = new StringBuffer("CREATE VIEW\n" + name + " AS\n");

    // generate select clause
    size = select.size();

    if(size > 0)
    {
      buf.append("SELECT\n");
      Enumeration enum = select.elements();
      for(int i = 0; i <size-1; i++)
        if(enum.hasMoreElements())
          buf.append(((Attribute)((Attribute_Component)enum.nextElement()).getAttribute()).getFullName() +",\n");

      buf.append(((Attribute)((Attribute_Component)enum.nextElement()).getAttribute()).getFullName()+"\n");
    }

    // generate from clause
    size = from.size();

    if(size > 0)
    {
      buf.append("FROM\n");
      Enumeration enum = from.elements();
      for(int i = 0; i <size-1; i++)
        if(enum.hasMoreElements())
          buf.append(((Relation_Component)enum.nextElement()).getRelation().getFullName() +",\n");

      buf.append(((Relation_Component)enum.nextElement()).getRelation().getFullName() +"\n");
    }

    if(where != null){
      // generate where clause
      size = where.size();

      if(size > 0)
      {
        buf.append("WHERE\n");
        Enumeration enum = where.elements();
        for(int i = 0; i <size-1; i++)
          if(enum.hasMoreElements())
            buf.append(((Condition_Component)enum.nextElement()).getCondition().toString()+"\n  AND ");

        buf.append(((Condition_Component)enum.nextElement()).getCondition().toString()+"\n");
      }
    }

    return buf.toString();
  }

  public int getVExtent(){
    return VExtent;
  }

  /**
  * Get clause from this view definition.
  * @param type type of clause. SELECT for SELECT clause, FROM for
  * FROM clause, WHERE for WHERE clause. These are constants in this class.
  * @return for type=SELECT: A vector of Attribute_Component.
  * @return for type=FROM  : A vector of Relation_Component.
  * @return for type=WHERE : A vector of Condition_Component.  */
  public Vector getClause(int type){

    Enumeration enum = null;

    switch(type){
    case SELECT:
      enum = select.elements();
      break;
    case FROM:
      enum = from.elements();
      break;
    case WHERE:
      if(where != null)
          enum = where.elements();
      break;
    default:
      if (DEBUG) System.out.println("View_Definition.getClause() Wrong type");
      return null;
    }

    Vector result= new Vector();
    if(enum == null)
        return result;

    while(enum.hasMoreElements())
      result.addElement(enum.nextElement());

    return result;
  }

  /**
  * Get PSJ List of this View Definition.
  * <ul>
  * <li>The projection list is same as select list</li>
  * <li>The select list is local condition from where list</li>
  * <li>The join list is join condition from where list</li>
  * </ul>
  * @param type type of PSJ list. PROJECT for PROJECT list, SELECT for SELECT list, JOIN for JOIN list.
  * @return a vector of PSJ list.
  */
  public Vector getPSJList(int type){
    Enumeration enum;
    Vector list = new Vector();

    switch(type){
    case PROJECT:
      enum = select.elements();
      while(enum.hasMoreElements())
        list.addElement(enum.nextElement());
      break;
    case SELECT:
      enum = where.elements();
      while(enum.hasMoreElements()){
        Condition c = ((Condition_Component)enum.nextElement()).getCondition();
        if(c instanceof Local_Condition)
          list.addElement(c);
      }
      break;
    case JOIN:
      enum = where.elements();
      while(enum.hasMoreElements()){
        Condition c = ((Condition_Component)enum.nextElement()).getCondition();
        if(c instanceof Join_Condition)
          list.addElement(c);
      }
      break;
    default:
      if (DEBUG) System.out.println("View_Definition.getPSJList() Wrong type");
      break;
    }

    return list;
  }


  /**
  * Return view name.
  */
  public String getName(){
    return name;
  }

  /**
  * Return vector of ISs of this VD. 
  * Return a vector of Information source names (in String format).
  */
  public Vector getISs(){
    Vector ISs = new Vector();
    Vector temp = new Vector(); // stores IS references.
    // Get the Relation names stored in from clause
    Enumeration enum = from.elements();

    // Check each Relation name in the FROM clause
    while(enum.hasMoreElements()){
      IS is = ((Relation_Component)enum.nextElement()).getRelation().getIS();

      if(!temp.contains(is))
	temp.addElement(is);
    }
    
    for(int i = 0; i < temp.size(); i++){
      ISs.addElement(((IS)temp.elementAt(i)).getName());
    }
    return ISs;
  }

  /**
  * Descide whether VD contains some IS
  * @param IS_name Information source name.
  * @return true: VD has; false: VD hasn't
  */
  public boolean contains(String IS_name){

    // Get the Relation names stored in from clause
    Enumeration enum = from.elements();

    // Check each Relation name in the FROM clause
    while(enum.hasMoreElements()){
      IS is = ((Relation_Component)enum.nextElement()).getRelation().getIS();
      if(is.getName().equals(IS_name))
	return true;
    }
    return false;
  }

  /**
  * Descide whether VD contains some Relation
  * @param IS_name Information source name.
  * @param Rel_name Relation name.
  * @return true: VD has; false: VD hasn't
  */
  public boolean contains(Relation rel){

    if(rel == null)
      return false;

    if(from == null)
      return false;

    // check FROM clause
    if(from_get(rel)!=null)
      return true;

    return false;
  }

  /**
   * Get a Relation Component for specific Relation.
   * @param rel the Relation
   * @return if found, return RC reference, otherwise, null.
   */
  public Relation_Component from_get(Relation rel){
    if(rel == null)
      return null;

    if(from == null)
      return null;

    // check FROM clause
    for(int i=0; i<from.size(); i++){
      Relation_Component rc = (Relation_Component)from.elementAt(i);
      if(rc.getRelation() == rel)
    	return rc;
    }

    return null;    
  }

  /**
  * Descide whether VD contains some Attribute.
  * @param IS_name Information source name.
  * @param Rel_name Relation name.
  * @param Attr_name Attribute name.
  * @return true: VD has; false: VD hasn't.
  */
  public boolean contains(Attribute attr){
    if(attr == null)
      return false;

    if(select == null)
      return false;

    // compose the whole name of attribute.
    String name = attr.getFullName();

    if (DEBUG) System.out.println("debug:View_Definition.contains(attr): " + name);

    // check SELECT clause
    if(get(select, attr)!=null)
      return true;

    if (DEBUG) System.out.println("debug:View_Definition.contains(attr): not in select clause.");

    // check WHERE clause
    if(where.getConditions(name)!=null)
      return true;

    if (DEBUG) System.out.println("debug:View_Definition.contains(attr): not in where clause.");

    return false;
  }

  /**
   * Debug Function show the View_Definition
   */
  public String toString(){
    return getESQL();
  }
  
  /* converts View Definition in the strongest E-SQL
   */
  public void makeStrongest(){
    
    //  System.out.println("Starting makeStrongest()");
    
    /* This function gets all the attributes from the select clause  */
    
    if (DEBUG) System.out.println("attr has "+select.size()+" Elements");
    
    for (Enumeration attr = select.elements() ; attr.hasMoreElements() ;) {
      Attribute_Component oneAttr = (Attribute_Component)attr.nextElement();
      if (DEBUG) System.out.println(oneAttr);
      Initialize init = new Initialize ();
      init.InitializeRAC (oneAttr,this);
        
    }// ends the for loop

    for (Enumeration cond = where.elements() ; cond.hasMoreElements() ;) {
      Condition_Component oneCond = (Condition_Component)cond.nextElement();
      if (DEBUG) System.out.println(oneCond);
      Initialize init = new Initialize ();
      init.InitializeRC (oneCond,this);

    } // ends the condition for loop

  }
  
  /**
   * Check if this view is evolvable or not for a specific relation.
   * It depends on theorem 1 in POC/SPOC paper. If this view is not
   * evolvable, then all affected components are both indispensable
   * and non-replaceable (false, false). We assume the View_Definition
   * is in the strongest format.
   */
  public boolean evolvable(Relation rel){
    if(from == null){
      if (DEBUG) System.out.println("View_Definition:evolvable FROM clause is empty!");
      return true;
    }
    
    // find the relation component of this relation.
    Relation_Component rc = from_get(rel);
    
    if(rc == null){
      if (DEBUG) System.out.println("View_Definition:evolvable \"" + rel.getName() + "\" is not in this view(" + name + ").");
      return true;
    }
    
    if(rc.Dispensable || rc.Replaceable)
      return true;
    else
      return false;
    
  }
  
  /**
   * Similar to evolvable for del_relation. Under same assumption,
   * we only need to check the FROM clause.
   */
  public boolean evolvable(Attribute attr){
    return evolvable(attr.getRelation());
  }
  
  /**
   * Clone the view definition.
   * Need to be implemented.
   */
  public Object clone(){
    View_Definition clone = new View_Definition();
    clone.name = name;
    clone.VExtent = VExtent;
    clone.from = (Vector)from.clone();//DSRG.VSA.VSA.deepClone(from);
    if(DEBUG)System.out.println("View_Definition.clone() from.");
    clone.select = (Vector)select.clone();//DSRG.VSA.VSA.deepClone(select);
    if(DEBUG)System.out.println("View_Definition.clone() select.");
    if(where != null)
      clone.where = (Condition_index)where.clone();//(Condition_index)where.deepClone();
    if(DEBUG)System.out.println("View_Definition.clone() where.");
   
    return clone;
  }

  /**
  * Replace relation at From clause.
  * @param r relation to be replaced.
  * @param s substitute relation.
  */
  public void replaceRelFrom(Relation r, Relation s){
    Relation_Component rc;
    
    int size = from.size();
    
    for(int i = 0; i < size; i ++){
      rc = (Relation_Component)from.elementAt(i);
      if(rc.getRelation() == r){
        // Replace it
        Relation_Component tmp = (Relation_Component)rc.clone();
        tmp.setRelation(s);
        from.setElementAt(tmp, i);
        return;
      }
    }
    
    if (DEBUG) System.out.println("View_Definition:repalceRelFrom relation " + r + " not found in from clause.");
  }
  
  /**
  * Drop a vector of components from one view definition.
  */
  public void dropComponents(Vector s){
    // scan SELECT clause
    Enumeration enum = select.elements();

    while(enum.hasMoreElements()){
      Attribute_Component ac = (Attribute_Component)enum.nextElement();

      if(s.contains(ac))
        select.removeElement(ac);
    }

    // scan FROM clause
    enum = from.elements();

    while(enum.hasMoreElements()){
      Relation_Component rc = (Relation_Component)enum.nextElement();

      if(s.contains(rc))
        from.removeElement(rc);
    }


    // scan WHERE clause
    if(where == null)
      return;

    for(int i = 0; i < where.size(); i++){
      if(s.contains(where.elementAt(i)))
        where.removeElementAt(i);
    }

    return;
  }
  /**
  * Get Dispensable components of one view defintion.
  */
  public Vector getDispensableComponent(){
    Vector result = new Vector();
    // Search SELECT clause
    Enumeration enum = select.elements();

    while(enum.hasMoreElements()){
      Attribute_Component ac = (Attribute_Component)enum.nextElement();

      if(ac.Dispensable)
        result.addElement(ac);
    }
    // Search WHERE clause
    if(where == null)
      return result;

    enum = where.elements();

    while(enum.hasMoreElements()){
      Condition_Component cc = (Condition_Component)enum.nextElement();

      if(cc.Dispensable)
        result.addElement(cc);
    }

    return result;
  }

    /**
  * Get Relation_Component for specific Relation.
  * @param r Reference of Relation in MKB
  * @return Relation Component in FROM clause of View; null not found.
  */
  Relation_Component getRelComponent(Relation r){
    Enumeration enum = from.elements();
    while(enum.hasMoreElements()){
      Relation_Component rc = (Relation_Component)enum.nextElement();

      if(rc.getRelation() == r)
        return rc;
    }
    return null;
  }

  /**
  * Get Condition_Component for specific Attribute.
  * @param a Reference of Attribute in MKB
  * @return a vector of Condition_Component which contains the attribute.
  */
  public Vector getCondComponents(Attribute a){
    if(where != null)
      return where.getCondComponents(a.getFullName());
    else
      return new Vector();
  }

  /**
  * Get an Attribute_Component from select clause for specific attribute
  * @param attr The reference of attribute.
  * @return the Attribute Component; not found return null
  */
  public Attribute_Component getAttrComponent(Attribute attr){
    return get(select, attr);
  }

  /**
  * Get an Attribute Component from a vector of attribute_components clause for specific Attribute.
  * @param vec the vector of Attribute Components.
  * @param attr The reference of attribute.
  * @return the Attribute Component; not found return null
  */
  Attribute_Component get(Vector sec, Attribute attr){
    Enumeration enum = sec.elements();

    while(enum.hasMoreElements()){
      Attribute_Component ac = (Attribute_Component)enum.nextElement();
      if(ac.getAttribute() == attr)
        return ac;
    }
    return null;
  }

  /**
  * Drop attribute-components from Select clause for specific vector of attributes
  * @param attrs Vector of attributes
  */
  public void dropAttrSelect(Vector attrs){
    Enumeration enum = attrs.elements();

    while(enum.hasMoreElements()){
      Attribute att = (Attribute)enum.nextElement();
      Attribute_Component ac = getAttrComponent(att);
      if(ac != null)
        select.removeElement(ac);
      else
        if (DEBUG) System.out.println("View_Definition:dropAttrSelect error: AC of attribute " + att.getFullName() + " not found!");
    }
  }

  /**
  * Add Relation to from clause of the View Definition.
  * @param r Reference of relation to be added.
  */
  public void addRelFrom(Relation r){
    if(getRelComponent(r) == null){
      Relation_Component rc = new Relation_Component(r, true, true);
      from.addElement(rc);
    }
  }

  /**
  * Add a Vector of JCs to the View Definition.
  * Here I haven't considering the correctness of Join_Conditions.
  * The default parameter for it is: true, true.
  * @param jcs A Vector of Join Conditions
  */
  public void addJCs(Vector jcs){
    Enumeration enum = jcs.elements();

    // Create Conditoin_Component for each Join_Conditions
    while(enum.hasMoreElements()){
      Join_Condition jc = (Join_Condition)enum.nextElement();
      if(!where.contains(jc)){
	Condition_Component cc = new Condition_Component(jc);
	where.addElement(cc);
      }
    }
  }

  /**
  * Replace attribute in the Where clause.
  * @param attr Attribute to be replaced.
  * @param reattr Substitute Attribute.
  */
  public void replaceAttrWhere(Attribute attr, Attribute reattr){

    if(where == null)
      return ;
    int size = where.size();

    // If we found the Condition, then we replace it.
    for(int i = 0; i < size; i++){
      Condition_Component cc = (Condition_Component)where.elementAt(i);
      Condition c = cc.getCondition();

      if(c.contains(attr.getFullName()) && cc.Replaceable == true){
        if(c.getLeft() == attr){
          // Replace it
          Condition_Component tmp = (Condition_Component)cc.clone();
          Condition tmp1 = tmp.getCondition();
          tmp1.setAttribute(reattr);
          where.setElementAt(tmp,i);
      	}
        if(c instanceof Join_Condition){
          Join_Condition jc = (Join_Condition) c;
          if(jc.getRight() == attr){
            // Replace it
            Condition_Component tmp = (Condition_Component)cc.clone();
            Join_Condition tmp1 = (Join_Condition)tmp.getCondition();
            tmp1.setRight(reattr);
            where.setElementAt(tmp, i);
      	  }
        }
      }else if(c.contains(attr.getFullName()) && cc.Replaceable == false){
        if (DEBUG) System.out.println("View_Definition:replaceAttrWhere: Error: Condition " + c + " (contians " + attr + ") cannot be replaced!");
      }
    }// endwhile
    return ;

  }


  /**
  * Get all components of one Relation
  * @param rel the specific Relation.
  * @return a Vector of components or a empty vector
  */
  public Vector getComponents(Relation rel){
    Vector result = new Vector();

    // Check Select Clause
    Enumeration enum = select.elements();
    while(enum.hasMoreElements()){
      Attribute_Component ac = (Attribute_Component)enum.nextElement();
      if(ac.getAttribute().getRelation() == rel)
        result.addElement(ac);
    }

    // Check From clause
    Relation_Component rc = (Relation_Component) from_get(rel);
    if(rc != null)
      result.addElement(rc);

    // Check Where clause
    if(where != null){
      Vector ccs = where.getCondComponents(rel);
      enum = ccs.elements();
      
      while(enum.hasMoreElements()){
        result.addElement(enum.nextElement());
      }
    }
    
    return result;
  }


  /**
  * Get all attributes of one Relation of this View. It will contains all the
  * attributes in the SELECT clause, and all the attributes in the JOIN Condition
  * of the WHERE clause.
  * @param rel the specific Relation.
  * @return a Vector of unique attributes or a empty vector
  */
  public Vector getAttributes(Relation rel){
    Vector result = new Vector();

    // Check Select Clause
    Enumeration enum = select.elements();
    while(enum.hasMoreElements()){
      Attribute_Component ac = (Attribute_Component)enum.nextElement();
      if(ac.getAttribute().getRelation() == rel)
        result.addElement(ac.getAttribute());
    }

    // Check Where clause
    if(where != null){
      Vector ccs = where.getCondComponents(rel);
      enum = ccs.elements();
      
      while(enum.hasMoreElements()){
        Condition c = ((Condition_Component)enum.nextElement()).getCondition();
        if(c instanceof Join_Condition){
            Join_Condition jc = (Join_Condition)c;
            Attribute left = jc.getLeft();
            Attribute right = jc.getRight();
            if(!result.contains(left))
                result.addElement(left);
            if(!result.contains(right))
                result.addElement(right);
        }
      }
    }
    return result;
  }

}

// EKTA Code start here.

class Initialize implements java.io.Serializable{
  
  public static void InitializeRAC (Attribute_Component attr, View_Definition that) {
    
    /** for each attribute , get the array value and then make a copy of it  **/
    /** similarly do for the relation and the condition  */
    Attribute_Component att = attr;
    //    if (DEBUG) System.out.println(att);
    Attribute attr1 = (Attribute)att.getAttribute();
    int avalue1, rvalue1, cvalue1,avalue2, rvalue2, cvalue2, av, rv, cv;
    Relation rel;
    Relation_Component rel_comp = new Relation_Component();
    ValueClassRA val = new ValueClassRA();
    ValueClassCA val1 = new ValueClassCA(); 
/** attr has the attribute components attr1 is the attribute object  **/
    
    
    avalue1 = (int)ArrayConvert.A_Bool_Int_Convert (att);
    //if (DEBUG) System.out.println("at comp" + avalue1);
       
    /** gets the relation for attribute   **/
    
    rel = (Relation)attr1.getRelation ();
    
    
    /** gets the components of the relation   **/
    
    rel_comp = (Relation_Component)that.getRelComponent(rel);
    //    if (DEBUG) System.out.println(rel_comp);
    rvalue1 = (int)ArrayConvert.R_Bool_Int_Convert(rel_comp);
    //    if (DEBUG) System.out.println("rel comp "  + rvalue1);
    
    /** this gets the integer values for the strongest ESQL for A, R  **/
    val = Compare.CompareRA (rvalue1, avalue1 );
    av = val.a1;
    //    if (DEBUG) System.out.println ("returned v.a " + av);
    
    rv = val.r1;
    //    if (DEBUG) System.out.println ("returned v.r " + rv);

    ArrayConvert.A_Int_Bool_Convert (att,av);
    ArrayConvert.R_Int_Bool_Convert (rel_comp,rv);
    
    /** gets the condition which has the the attributes   **/
    /** there can be more than one condition that has that attribute  **/
    
      
    Vector v = that.getCondComponents(attr1);
    
    for (int i = 0; i < v.size() ; i++) {
      //      if (DEBUG) System.out.println("i is " +i); 
      Condition_Component cc = (Condition_Component)v.elementAt(i);
      //      if (DEBUG) System.out.println("condition com is " + cc);
      cvalue1 = (int)ArrayConvert.C_Bool_Int_Convert(cc);
      //      if (DEBUG) System.out.println(cvalue1);
         
    
      /** this gets the integer values for the strongest ESQL for A, C  **/
      val1 = Compare.CompareCA (cvalue1, avalue1);
      av = val1.a1;
      //      if (DEBUG) System.out.println ("returned v.a " + av);
      cv = val1.c1;
      //      if (DEBUG) System.out.println ("returned v.c is " + cv);
      
      /** this converts the integer values back to the boolean values  **/
    
      ArrayConvert.A_Int_Bool_Convert (att,av);
      ArrayConvert.C_Int_Bool_Convert (cc,cv);
      
    }
  }

  public static void InitializeRC (Condition_Component cond, View_Definition that) {
    int rvalue1, cvalue1, av, rv, cv;
    ValueClassRC val3 = new ValueClassRC();
    cvalue1 = (int)ArrayConvert.C_Bool_Int_Convert(cond);

    Vector v = cond.getRelations();
    
    for (int i = 0; i < v.size() ; i++) {
      //      if (DEBUG) System.out.println("i is " +i); 
      Relation rel = (Relation)v.elementAt(i);
      //      if (DEBUG) System.out.println("relation is " + rel);
      Relation_Component rel_comp = that.getRelComponent(rel);
      rvalue1 = (int)ArrayConvert.R_Bool_Int_Convert(rel_comp);
      
      //      cvalue1 = (int)ArrayConvert.C_Bool_Int_Convert(cc);
      //      if (DEBUG) System.out.println(cvalue1);
	         
      val3 = Compare.CompareRC (rvalue1, cvalue1);
      rv = val3.r1;
      //      if (DEBUG) System.out.println ("returned v.r " + rv);
      cv = val3.c1;
      //      if (DEBUG) System.out.println ("returned v.c is " + cv);
      
      /** this converts the integer values back to the boolean values  **/
    
      ArrayConvert.R_Int_Bool_Convert (rel_comp,rv);
      ArrayConvert.C_Int_Bool_Convert (cond,cv);
    }
  }
  
}

/** this class is mainly used to return more than one value
    if any function needs to  **/
/** eg. the compare function needs to return all three values for attr, rel
    and cond  **/

class ValueClassRA implements java.io.Serializable {

    int a1, r1;
    public ValueClassRA () {
    }
}

class ValueClassCA implements java.io.Serializable {

    int c1,a1;
    public ValueClassCA () {
    }
}

class ValueClassRC implements java.io.Serializable {

    int r1,c1;
    public ValueClassRC () {
    }
}




/** this class is used to convert the boolean values of Replacable and
    Dispensible to integer values   **/
/** it is also used to convert the integer values back to boolean values  **/


class ArrayConvert implements java.io.Serializable {

public static int A_Bool_Int_Convert(Attribute_Component comp) {

    int value=0;
    
    if ((comp.Dispensable == false) && (comp.Replaceable == false))
        value = 0;

    if ((comp.Dispensable == false) && (comp.Replaceable == true))
        value = 1;

    if ((comp.Dispensable == true) && (comp.Replaceable == false))
        value = 2;

    if ((comp.Dispensable == true) && (comp.Replaceable == true))
        value = 3;

    return (value);

  }            // ends the function "Bool_int_convert"


public static int R_Bool_Int_Convert(Relation_Component comp) {

    int value=0;
    
    if ((comp.Dispensable == false) && (comp.Replaceable == false))
        value = 0;

    if ((comp.Dispensable == false) && (comp.Replaceable == true))
        value = 1;

    if ((comp.Dispensable == true) && (comp.Replaceable == false))
        value = 2;

    if ((comp.Dispensable == true) && (comp.Replaceable == true))
        value = 3;

    return (value);

  }            // ends the function "R_Bool_int_convert"


public static int C_Bool_Int_Convert(Condition_Component comp) {

    int value=0;
    
    if ((comp.Dispensable == false) && (comp.Replaceable == false))
        value = 0;

    if ((comp.Dispensable == false) && (comp.Replaceable == true))
        value = 1;

    if ((comp.Dispensable == true) && (comp.Replaceable == false))
        value = 2;

    if ((comp.Dispensable == true) && (comp.Replaceable == true))
        value = 3;

    return (value);

  }            // ends the function "C_Bool_int_convert"



public static void R_Int_Bool_Convert (Relation_Component rel_comp, int value1) {
    int value = value1;
    //    if (DEBUG) System.out.println ("now in rel int bool convert");    
    if (value == 0) {
        rel_comp.Dispensable = false;
        rel_comp.Replaceable = false;
    }


    if (value == 1) {
        rel_comp.Dispensable = false;
        rel_comp.Replaceable = true;
    }


    if (value == 2) {
        rel_comp.Dispensable = true;
        rel_comp.Replaceable = false;
    }


    if (value == 3) {
        rel_comp.Dispensable = true;
        rel_comp.Replaceable = true;
    }

    //    if (DEBUG) System.out.println ("new relation dis is "+ rel_comp);
  

  }    // ends function "Int_Bool_Convert

public static void A_Int_Bool_Convert (Attribute_Component att, int value1) {
  //  if (DEBUG) System.out.println ("now in A int bool convert");
    int value = value1;
    //    if (DEBUG) System.out.println ("now in att int bool convert");    
    
    if (value == 0) {
        att.Dispensable = false;
        att.Replaceable = false;
	//	if (DEBUG) System.out.println ("if value is 0");
    }
    else {

      if (value == 1) {
        att.Dispensable = false;
        att.Replaceable = true;
	//	if (DEBUG) System.out.println ("if value is 1");
      }
      else {

	if (value == 2) {
	  att.Dispensable = true;
	  att.Replaceable = false;
	  //	  if (DEBUG) System.out.println ("if value is 2");
	}
	else {

	  if (value == 3) {
	    att.Dispensable = true;
	    att.Replaceable = true;
	    //	    if (DEBUG) System.out.println ("if value is 3");
	  }
	}
      }
    }
    //    if (DEBUG) System.out.println ("att is" + att);    
  
  }    // ends function A_Int_Bool_Convert

public static void C_Int_Bool_Convert (Condition_Component cc,int value) {

    if (value == 0) {
        cc.Dispensable = false;
        cc.Replaceable = false;
    }


    if (value == 1) {
        cc.Dispensable = false;
        cc.Replaceable = true;
    }


    if (value == 2) {
        cc.Dispensable = true;
        cc.Replaceable = false;
    }


    if (value == 3) {
        cc.Dispensable = true;
        cc.Replaceable = true;
    }

  }    // ends function C_Int_Bool_Convert

}       // ends the class "arrayconvert



class Compare implements java.io.Serializable {
  
  public static ValueClassRA CompareRA (int rr1,int aa1) {	
    //    if (DEBUG) System.out.println("comparing RA");
    ValueClassRA v = new ValueClassRA();
    int aa2, rr2, arr;
    
    ArrayStore rac = new ArrayStore ();
    
    do {
	  
      aa2=aa1;
      //      if (DEBUG) System.out.println("aa2 is " + aa2);
      rr2=rr1;
      //      if (DEBUG) System.out.println("rr2 is" + rr2);
         
      /**  This compares the relation and the attribute and returns the array value **/
      
      arr = (int)rac.StoreValuesRA(rr1,aa1);
      //      if (DEBUG) System.out.println ("storevalues " + arr);	
      if (arr > 8) {
        aa1 = arr - 9;
	//	if (DEBUG) System.out.println ("now aa1 " + aa1);
      }
      else {
	if (arr > 4) {
	  rr1 = arr - 5;
	  //	  if (DEBUG) System.out.println ("now rr1 " + rr1);
	}
      }
      //      if (DEBUG) System.out.println ("if no values then no change");
      
      //      if (DEBUG) System.out.println ("before ending while loop");
      
    }while ((aa1!=aa2) || (rr1 != rr2));
    
    /** this compares if there is any change in the values since the last loop **/
    
    v.a1 = aa1;
    //    if (DEBUG) System.out.println ("v.aa1 " + aa1);
    v.r1 = rr1;
    //    if (DEBUG) System.out.println ("v.rr1 " + rr1);
    return (v);
     
  }

 public static ValueClassCA CompareCA (int cc1,int aa1) {	
   //    if (DEBUG) System.out.println("comparing CA");
    ValueClassCA v = new ValueClassCA();
    int aa2, cc2, arr;
    ArrayStore rac = new ArrayStore ();
    
    do {
	  
      aa2=aa1;
      //      if (DEBUG) System.out.println("aa2 is " + aa2);
      cc2=cc1;
      //      if (DEBUG) System.out.println("cc2 is " + cc2);
      
      /** This compares the attribute and the condition  **/
      
      arr = (int)rac.StoreValuesCA(cc1,aa1);
      if (arr > 8){
        aa1 = arr - 9;
	//	if (DEBUG) System.out.println ("aa1 now is " + aa1);
      }
      else {
	if (arr > 0) {
	  cc1 = arr - 1;
	  //	  if (DEBUG) System.out.println ("cc1 now is " + cc1);
	}
      }

      //      if (DEBUG) System.out.println ("before ending while loop");
      
    }while ((aa1!=aa2) || (cc1 != cc2));
    
    /** this compares if there is any change in the values since the last loop **/
    
    v.a1 = aa1;
    //    if (DEBUG) System.out.println ("v.aa1 " + aa1);
    v.c1 = cc1;
    //    if (DEBUG) System.out.println ("v.cc1 " + cc1);
    return (v);
    
    
  }
  
 public static ValueClassRC CompareRC (int rr1,int cc1) {	
   //    if (DEBUG) System.out.println("comparing RC");
    ValueClassRC v = new ValueClassRC();
    int rr2, cc2, arr;
    ArrayStore rac = new ArrayStore ();
    
    do {
	  
      rr2=rr1;
      //      if (DEBUG) System.out.println("rr2 is" + rr2);
      cc2=cc1;
      //      if (DEBUG) System.out.println("cc2 is " + cc2);
      
      /** This compares the relation and the condition  **/
      
      arr = (int)rac.StoreValuesRC(rr1,cc1);
      if (arr > 4){
        rr1 = arr - 5;
	//	if (DEBUG) System.out.println ("rr1 now is" + rr1);
      }
      else {
	if (arr > 0) {
	  cc1 = arr - 1;
	  //	  if (DEBUG) System.out.println ("cc1 now is " + cc1);
	}
      }
      //      if (DEBUG) System.out.println ("if no values then no change");

      //      if (DEBUG) System.out.println ("before ending while loop");
      
    }while ((rr1 != rr2) || (cc1 != cc2));
    
    /** this compares if there is any change in the values since the last loop **/
    
    v.r1 = rr1;
    //    if (DEBUG) System.out.println ("v.rr1 " + rr1);
    v.c1 = cc1;
    //    if (DEBUG) System.out.println ("v.cc1 " + cc1);
    return (v);
    
  }
  
}


class ArrayStore implements java.io.Serializable {

/* This class is used to store the different values in the tables used for
the strongest ESQL conversion */
/* if there is no change in any value, the array value stored is 0 */


/*      Value 0 is stored incase there is no change */
/*      value 1-4 is for Condition        */
/*      Value 5-8 is for Relation     */
/*      Value 9-12 is for Attribute     */


public static int StoreValuesRA (int a, int b) {
  // if (DEBUG) System.out.println ("Running StorevaluesRA");
  int i, j, ra;
  int[][] RA = new int [4][4];


     for ( i = 0 ; i < 4 ; i ++) {
                for ( j = 1 ; j < 4 ; j++ ) {
                      RA [i][j] = 0;
                                            
                 }
        }

        RA [0][0] = 0;
        
        for ( i = 1 ; i < 4 ; i ++ ) {
             RA [i][0] = 5;
                     
        }

	ra = RA[a][b];
	return ra;
        

}                       

  public static int StoreValuesRC (int a, int b) {
    //    if (DEBUG) System.out.println ("running storevaluesRC");     
    int i, j, rc;
    int[][] RC = new int [4][4];
    
    for ( i = 1 ; i < 4 ; i ++) {
      for ( j = 0 ; j < 4 ; j++ ) {
	RC [i][j] = 0;
	
      }
    }
    
    RC [0][0] = 0;
    
    for ( i = 1 ; i < 4 ; i ++ ) {
      RC [i][0] = 5;
      
    }
    
    rc = RC[a][b];
    return (rc);
    
  }                       
  
  public static int StoreValuesCA (int a, int b) {
    //    if (DEBUG) System.out.println ("Running StorevaluesCA");
    int i, j, ca;
    int[][] CA = new int [4][4];


    for (i = 1 ; i < 4 ; i ++) {
      CA [i][0] = 1;      
    }
    

    for (j = 1 ; j < 4 ; j ++) {              
      CA [0][j] = 9;      
    }
    

    for (i = 1 ; i < 4 ; i ++) {
      CA [i][2] = 0;      
    }
    

    for (i = 1 ; i < 3 ; i ++) {
      CA [i][1] = 0;      
    }
    

    CA [0][0] = 0;
    CA [1][3] = 10;
    CA [3][3] = 0;
    CA [3][1] = 2;
    CA [2][3] = 0;
    

    ca = CA[a][b];


    return (ca);    
  }
}


