////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.VSA;

import DSRG.VKB.VKB;
import DSRG.VKB.View_Definition;
import DSRG.VKB.Attribute_Component;
import DSRG.VKB.Join_Condition;
import DSRG.VKB.Condition_Component;
import DSRG.VKB.View_Component;
import DSRG.VKB.Version;
import DSRG.VKB.QC;


import DSRG.MKB.MKB;
import DSRG.MKB.Relation;
import DSRG.MKB.Attribute;


import java.util.Enumeration;
import java.util.Vector;

import java.lang.reflect.*;

/**
* View Synchronization Algorithm of DSRG. It only handle the View
* Definition Evolution.  <p> It includes following functions: <ul>
* <li>Add Attribute</li> <li>Add Relation</li> <li>Delete
* Attribute</li> <li>Delete Relation</li> <li>Change Attribute
* Name</li> <li>Change Relation Name</li> </ul> Add Attribute and Add
* Relation is the responsibility of MKB.  <p> Assume VKB will only
* represents one Data warehouse. But, VSA can apply to all the data
* warehouses. Here, I make the VSA class static, it will just provide
* the algorithm or functions to evolve the VKB and MKB. In later, we
* maybe create a super class EVE which will contain the VKB and MKB
* objects.  <p> VSA will give out two view definitions. VA (or View
* Evolver) will adapt the view from the two view definitions.  */

public class VSA implements java.io.Serializable
{
    /**
     * The MKB this VSA working on. That's the place where VSA can get
     * meta-knowledge.  */
    MKB mkb;
    
    /**
     * The VKB this VSA working on. That's the object VSA will
     * modify. It stores view definitions in an internal format.  */
    VKB vkb;
    
    /** 
     * the tradeoff parameter for the attributes (therefore also for
     * the extent). Used for creating QC-Objects */
    private double rho_attr;

    /** 
     * the tradeoff parameter for the quality (therefore also for the
     * cost). Used for creating QC-Objects */
    private double rho_quality;

    /** the tradeoff parameters for the costs (unit costs) */
    private double cost_M, cost_T, cost_IO;


    /**
     * Constructor of VSA module.
     */
    public VSA(MKB m, VKB v){
        
    	mkb = m;
    	vkb = v;
        rho_attr = 
            java.lang.Double.valueOf(System.getProperty("rho_attr")).doubleValue();
        rho_quality = 
            java.lang.Double.valueOf(System.getProperty("rho_quality")).doubleValue();
        
        cost_M = java.lang.Double.valueOf(System.getProperty("cost_M")).doubleValue();
        cost_T = java.lang.Double.valueOf(System.getProperty("cost_T")).doubleValue();
        cost_IO = java.lang.Double.valueOf(System.getProperty("cost_IO")).doubleValue();

    }
    
    /**
     * Constructor of VSA module with setting the systemwide tradeoff
     * parameters.  */
    /*
    public VSA(MKB m, VKB v, double my_rho_attr, double my_rho_quality){
    	mkb = m;
    	vkb = v;
    	rho_attr = my_rho_attr;
    	rho_quality = my_rho_quality;
    }
    */

    /**
     * This method will be called when there is a Schema_Change collected
     * by GISL. It will consume one schema evolution and evolve the view.
     * <p> This method will give out two view definitions. One if orginal
     * view definition, the other one is evolved view definition.  <p> It
     * will do three steps: <ol> <li> Get Schema_Change from GISL <li>
     * Depends on Schema_Change, get affected View Definitions from VKB.
     * <li> for each View Definition, do VSA (delete , add,
     * Schema_Change)(rel, attr) </ol>
     * @param chg Schema_Change which comes from Global IS Layer (GISL), which
     * will collect the update from IS and send them to VSA.  */
    
    public void viewEvolution(Schema_Change chg){
	
	try{
	    Vector affected_views; // affected View definitions.
	    Vector evV; // evolvable view definitions.
	    Vector obV; // NOT evolvable view definitions.
	    
	    // get Schema_Change
	    if(chg == null){
		System.out.println("VSA.viewEvolution: null Schema_Change");
		return;
	    }
	  
	    switch(chg.type){
	    case Schema_Change.CHG_REL:
		mkb.changeRelationName(chg.IS_name, chg.Rel_name, chg.new_name);
		System.out.println("VSA.viewEvolution: in MKB " + chg.IS_name + "." 
				   + chg.Rel_name + " --> " + chg.IS_name + "." + chg.new_name);
		return;
	    case Schema_Change.CHG_ATT:
		mkb.changeAttributeName(chg.IS_name, chg.Rel_name, chg.Attr_name, chg.new_name);
		System.out.println("VSA.viewEvolution: in MKB "
				   + chg.IS_name + "." + chg.Rel_name + "." + chg.Attr_name
				   + " --> "
				   + chg.IS_name + "." + chg.Rel_name + "." + chg.new_name);
		return;
	    case Schema_Change.ADD_ATT:
	    case Schema_Change.ADD_REL:
		System.out.println("VSA.viewEvolution: not affected");
		return;
	    case Schema_Change.DEL_REL:
	    case Schema_Change.DEL_ATT:
		Attribute att = null;
		Relation rel = null;
		affected_views = null;
	
		if(chg.type == Schema_Change.DEL_ATT){
		    affected_views = (Vector)vkb.getViewDefinitions(chg.IS_name, chg.Rel_name, chg.Attr_name);
		  
		    att = mkb.findAttribute(chg.IS_name, chg.Rel_name, chg.Attr_name);
		}else if(chg.type == Schema_Change.DEL_REL){
		    affected_views = (Vector) vkb.getViewDefinitions(chg.IS_name, chg.Rel_name);
		  
		    rel = mkb.findRelation(chg.IS_name, chg.Rel_name);
		}
	      
		if(affected_views == null){
		    System.out.println("VSA.viewEvolution: no view affected");
		    return;
		}
	      
		// 1. Find set of evV (evolvable affected Views)
		// 2. Find set of obV (NOT evolvable affected Views)
		evV = new Vector();
		obV = new Vector();
		Enumeration enum = affected_views.elements();
		while(enum.hasMoreElements()){
		    View_Definition vd = (View_Definition)enum.nextElement();
		    if(chg.type == Schema_Change.DEL_ATT){
			if(vd.evolvable(att))
			    evV.addElement(vd);
			else
			    obV.addElement(vd);
		    }else if(chg.type == Schema_Change.DEL_REL){
			if(vd.evolvable(rel))
			    evV.addElement(vd);
			else
			    obV.addElement(vd);
		    }
		}
		// 3. Delete obV from VKB
		vkb.dropSetOfVDs(obV);
		// 4. For each V in evV DO
		enum = evV.elements();
		while(enum.hasMoreElements()){
		    View_Definition V = (View_Definition)enum.nextElement();
		    // 5.   rewritingsV = POC(ch, V)
		    Vector rewritingsV = null;
		    if(chg.type == Schema_Change.DEL_ATT)
			rewritingsV = POC(V, att);
		    else if(chg.type == Schema_Change.DEL_REL)
			rewritingsV = POC(V, rel);
		  
		    // if no legal rewritings, delete it, and go to next view.
		    if(rewritingsV == null){
			vkb.dropViewDefinition(V);
			continue;
		    }
		  
		    // 6.   rewritingsV = rewritingsV union ComputeDerived(rewritingsV)
		    rewritingsV = computeDerived(rewritingsV);
		  
		    //if(DSRG.MW.Main.DEBUG)
		    //System.out.println("Debug: VSA.viewEvolution: derived views: " + rewritingsV);


		    /**
		     * For QC model need to interact with User, VSA will not directly update the VKB.
		     * Instead it will put the different versions into that View_Definition.
		     // 7.   IF rewritingsV = empty
		     if(rewritingsV.isEmpty())
		     // 8.     THEN delete V from VKB
		     vkb.dropViewDefinition(V.getName());
		     else{
		     // 9.     ELSE V' = the best from rewritingsV
		     View_Definition bestV = findBest(rewritingsV, V);
		     // 10.         replace V with V' in VKB
		     vkb.replaceViewDefinition(V, bestV);
		     }
		     }
		    */

		    // 7. Put the rewritings into the VKB.
		    if(!rewritingsV.isEmpty()){
			int size = rewritingsV.size();
			Vector vs = new Vector();
                        // reset max. Cost (for normalization)
                        DSRG.VKB.QC.resetMaxCost();
                        // remove duplicates
			for(int i = 0; i < size; i++){
                            boolean already_there = false;
                            View_Definition vd = (View_Definition)rewritingsV.elementAt(i);
                            for (int j = 0; j < vs.size(); j++) {
                                if (((Version)vs.elementAt(j)).getDefinition()
                                    .equals(vd.getSQL()))
                                    already_there = true;
                            }
                            if (!already_there) {
                                Version tempVersion = new 
                                    Version(V, vd, 
                                            new QC(rho_attr, rho_quality, 
                                                   cost_M, cost_T, cost_IO),mkb);
                                vs.addElement(tempVersion);
                            }
			}
                        vs = dumbSort(vs);
			V.setVersions(vs);
		    }else{
			vkb.dropViewDefinition(V.getName());
		    }
		    return;
		}
	    default:
		System.out.println("VSA.viewEvolution: wrong type of Schema_Change");
		return;
	    }
	}catch(Exception e){
	    System.out.println("debug:VSA.viewEvolution Exception : " + e.getMessage());
	    e.printStackTrace();
	}
    }
    
    /**
     * Handle Schema Change: Delete Relation by using POC algorithm.
  * @param vd View defintion which is going to be evolved.
  * @param rel Relation in MKB which will be dropped.
  * @return Basic Legal rewritings of vd. It cannot be null.
  */
    Vector POC(View_Definition vd, Attribute att){
    
	Vector atleast = new Vector(); // Attributes that at least appeared in the upperline{A}.
	Relation rel = att.getRelation(); // Relation R of R.A
	// Attribute of Select Clause.
	Vector asaa = union(
			    union((Vector)vd.attrSelect(rel, true, true),
				  (Vector)vd.attrSelect(rel, false, true)),
			    union((Vector)vd.attrSelect(rel, true, false),
				  (Vector)vd.attrSelect(rel, false, false)));

	// Attribute of Attr(V(R))(*, f)
	Vector awaf = union((Vector)vd.attrWhere(rel, true, false),
			    (Vector)vd.attrWhere(rel, false, false));

	// Attribute of Attr(V(R))where
	Vector awaa = union(
			    union(awaf, (Vector)vd.attrWhere(rel, true, true)),
			    (Vector)vd.attrWhere(rel, false, true));

	Vector temp;
	Vector candidates; // Candidates of replacement of POC
	Vector attrs = null; // upperline(A)

	// 1. IF VE is SUBSET or EQUIVALENT
	if(vd.getVExtent() == View_Definition.SUBSET || vd.getVExtent() == View_Definition.EQUIVALENT){
	    // 2. if R.A belongs to Attr(V(R))where(*,f)
	    if(awaf.contains(att)){
		// Cannot Evolved
		System.out.println("VSA:POC View(" + vd.getName() + ") cannot be evolved.");
		return null;
	    }
  
	    // 3. Attr(V(R))where is subset of upperline{A}
	    atleast = (Vector)awaa.clone();

	    Vector asta = union((Vector)vd.attrSelect(rel, true, true), 
				(Vector)vd.attrSelect(rel, true, false));

	    temp = substract(asaa, awaa);

	    if(!subset(temp, asta)){
		// 4. if temp is not a subset of asta, then cannot evolved
		System.out.println("VSA:POC View(" + vd.getName() + ") cannot be evolved.");
		return null;
	    }
	}
	// 5. Prepare VE and a vector of attributes including R.A
	temp = new Vector();
	temp.addElement(att);
	atleast = union(atleast, temp);

	// 6. MKB Search Pi(upperline{B}(S)) phi Pi(upperline{A}(R))
	candidates = (Vector)mkb.findPOC(rel, atleast, vd.getVExtent());
    
	Vector rewritingsV = new Vector();// Vector of rewritings

	// 6.5 If we cannot find any PC constraint
	if(candidates.size() == 0){
	    View_Definition rewrite = (View_Definition) vd.clone();
	    Vector drop = new Vector();
	    // Test if the Componenents of this Attribute is dispensable
	    Attribute_Component ac = rewrite.getAttrComponent(att);
	    if(ac != null){
		drop.addElement(ac);
		if(ac.Dispensable == false){
		    System.out.println("VSA:POC View(" + rewrite.getName() + ") cannot be evolved.");
		    return null;
		}
	    }

	    Vector ccs = rewrite.getCondComponents(att);
	    if(ccs != null){
		for(int i = 0; i< ccs.size(); i++){
		    Condition_Component cc =(Condition_Component) ccs.elementAt(i);
		    drop.addElement(cc);
		    if(cc.Dispensable == false){
			System.out.println("VSA:POC View(" + rewrite.getName() + ") cannot be evolved.");
			return null;
		    }
		}
	    }

	    // Drop all the Components of that Attribute.
	    rewrite.dropComponents(drop);

	    rewritingsV.addElement(rewrite);
	    return rewritingsV;
	}
        
	// 7. For each item in replacement (S, upperline{A} -> upperline{B})
	Enumeration enum = candidates.elements();

	while(enum.hasMoreElements()){
	    // 8. Generate basic legal rewritings for each (S, upperline{a} -> upperline{B})
	    Replacement rp = (Replacement) enum.nextElement();

	    View_Definition rewrite = (View_Definition) vd.clone();

	    // 9. IF R.A replaceable
	    Attribute_Component ac = vd.getAttrComponent(att);

	    if(ac.Replaceable){
		// 10. Replace S.B <-> R.A
		Attribute reatt = (Attribute)rp.attr_pairs.get(att);
		rewrite.replaceAttrSelect(att, reatt);
	    }

	    // Get the upperline{A} from the replacement
	    attrs = createVector(rp.attr_pairs.keys());

	    // if VE is SUBSET or EQUIVALENT
	    if(vd.getVExtent() == View_Definition.SUBSET || vd.getVExtent() == View_Definition.EQUIVALENT){
		temp = substract(asaa, attrs);
		// Drop (Attr(V(R))select - upperline(A))
		rewrite.dropAttrSelect(temp);
	    }
      
	    // Drop R.A from select clause
	    temp = new Vector();
	    temp.addElement(att);
	    rewrite.dropAttrSelect(temp);

	    // Add S to From clause
	    rewrite.addRelFrom(rp.rel);


	    // Generate Join Condition
	    // a. Calculate Attr(V(R))
	    Vector aaaa = union(awaa, asaa);
	    // b. Interset Attr(V(R)) and upperline(A)
	    Vector jcleft = intersect(aaaa, attrs);
	    // c. Substract Attr R.A from the Left
	    jcleft = substract(jcleft, temp);
	    // d. Generate a Vector of Join Condition
	    Vector jcs = new Vector();
	    for(int i = 0; i< jcleft.size(); i++){
		jcs.addElement(new Join_Condition((Attribute)jcleft.elementAt(i), (Attribute)rp.attr_pairs.get(jcleft.elementAt(i))));
	    }

	    // Add JC to rewrite
	    rewrite.addJCs(jcs);

	    // Replace R.A <-> S.B in Where clause.
	    rewrite.replaceAttrWhere(att, (Attribute)rp.attr_pairs.get(att));

	    rewritingsV.addElement(rewrite);

	}

	return rewritingsV;
    }


    /**
     * sorts a vector of Version objects by their QC-Values. Uses
     * simple simpleSort (n^2 complexity) */

    Vector dumbSort (Vector v) {
        if (v.size()>100) 
            System.out.println("Sorting "+v.size()+" Versions by QC-Value");
        Version one, two, temp;
        for (int i = 0 ; i < v.size() ; i++)
            for (int j = i + 1 ; j < v.size() ; j++) {
                one = (Version)v.elementAt(i);
                two = (Version)v.elementAt(j);
                if (
                    one.getQC().getQCValue() <
                    two.getQC().getQCValue() 
                    ) {
                    temp = one;
                    v.setElementAt(two,i);
                    v.setElementAt(temp,j);
                }
            }
        return v;
    }
        
    /**
  * Intersect two Vectors
  * @param vec1 First Vector
  * @param vec2 Second Vector
  * @return An intersection of two vectors. If that is empty return empty vector.
  */
    Vector intersect(Vector vec1, Vector vec2){
	Vector result = new Vector();

	if(vec1 == null || vec2 == null)
	    return result;

	Enumeration enum = vec1.elements();

	while(enum.hasMoreElements()){
	    Object elm = enum.nextElement();
	    if(vec2.contains(elm))
		result.addElement(elm);
	}

	return result;
    }

    /**
  * Create a Vector from Enumeration
  * @param enum Enumeration
  * @return a Vector, if enum is null, return empty vector
  */
    Vector createVector(Enumeration enum){
	Vector result = new Vector();

	if(enum == null)
	    return result;
  
    
	while(enum.hasMoreElements())
	    result.addElement(enum.nextElement());
      
	return result;
    }
  
    /**
  * Substract one Vector from another Vector.
  * @param sdv Substracted Vector
  * @param sv Substract Vector
  * @return Vector of result. or an empty vector.
  */
    Vector substract(Vector sdv, Vector sv){
	// For each element in SV, remove it from vector sdv.
	Vector temp = (Vector)sdv.clone();
	Enumeration enum = sv.elements();
	while(enum.hasMoreElements()){
	    Object elm = enum.nextElement();
	    temp.removeElement(elm);
	}
	return temp;
    }

    /**
  * Test if the first one it the subset of the second Vector
  * @param sbv Subset Vector
  * @param fsv FullsetVector
  * @return if true, means sbv is subset of fsv.
  */
    public static  boolean subset(Vector sbv, Vector fsv){
	// For each element in sbv, test if that is in the fsv
	Enumeration enum = sbv.elements();
	while(enum.hasMoreElements()){
	    Object elm = enum.nextElement();
	    if(!fsv.contains(elm))
		return false;
	} 
	return true;
    }

    /**
  * Handle Schema Change: Delete Relation by using POC algorithm.
  * @param vd View defintion which is going to be evolved.
  * @param rel Relation in MKB which will be dropped.
  * @return Basic Legal rewritings of vd. It cannot be null.
  */
    Vector POC(View_Definition vd, Relation rel){
    
	Vector atleast;// Attributes that at least appeared in the upperline{A}.
	Vector candidates;// stores Vector of S and Hashtable of upperline{A} -> upperline{B}.
	// 1. IF VE == SUPERSET
	if(vd.getVExtent() == View_Definition.SUPERSET){
	    // 2. IF Attr(V(R))where(t,f) != null
	    Vector awtf = (Vector)vd.attrWhere(rel, true, false); 
	    if(awtf != null){
		// 3. Cannot Evolved
		System.out.println("VSA:POC View(" + vd.getName() + ") cannot be evolved.");
		return null;
	    }
	    // 4. Get Attr(V(R))where(*, t)
	    atleast = union((Vector)vd.attrWhere(rel, true, true),
			    (Vector)vd.attrWhere(rel, false, true));
	}else{
	    // 5. Get Attr(V(R))where(f, t)
	    atleast = (Vector)vd.attrWhere(rel, false, true);
	}

	// 6.  union Attr(V(R))select(f,t) and R and phi
	atleast = union(atleast, (Vector)vd.attrSelect(rel, false, true));

	// 7. MKB Search Pi(upperline{B}(S)) phi Pi(upperline{A}(R))
	candidates = (Vector)mkb.findPOC(rel, atleast, vd.getVExtent());

	Vector rewritingsV = new Vector();// Vector of rewritings

	// 7.5 IF no PC Constraints Found
	if(candidates.size() == 0){
	    View_Definition rewrite = (View_Definition) vd.clone();
	    // Check if all components of this Relation is dispensable
	    Vector cs = rewrite.getComponents(rel);
	    Enumeration enum = cs.elements();
	    while(enum.hasMoreElements()){
		if(((View_Component)enum.nextElement()).Dispensable == false){
		    System.out.println("VSA:POC View(" + vd.getName() + ") cannot be evolved.");
		    return null;
		}
	    }

	    // Drop all the components
	    rewrite.dropComponents(cs);
	    rewritingsV.addElement(rewrite);
	    return rewritingsV;
	}
	// 8. For each item in replacement (S, upperline{A} -> upperline{B})
	Enumeration enum = candidates.elements();

	while(enum.hasMoreElements()){
	    // 9. Generate basic legal rewritings for each (S, upperline{a} -> upperline{B})
	    Replacement rp = (Replacement) enum.nextElement();

	    View_Definition rewrite = (View_Definition) vd.clone();

	    // 9.1 Drop attr(V(R))select/where(t,f)
	    rewrite.dropAttrSelect(rel, true, false);
	    rewrite.dropCondWhere(rel, true, false);

	    // 9.2 Replace attr(V(R))select/where(f,t);
	    if(!rewrite.replaceAttrSelect(rel, rp, false, true)){
		System.out.println("DEBUG: VSA.viewEvolution view: " + vd.getName() + " cannot be evolved");
		continue;
	    }
	    if(!rewrite.replaceCondWhere(rel, rp, false, true)){
		System.out.println("DEBUG: VSA.viewEvolution view: " + vd.getName() + " cannot be evolved");
		continue;
	    }

	    // 9.3 Replace attr(V(R))select/where(t,t);
	    rewrite.replaceAttrSelect(rel, rp, true, true);
	    rewrite.replaceCondWhere(rel, rp, true, true);

	    // 9.4 Replace relation R in the from clause;
	    rewrite.replaceRelFrom(rel, rp.rel);

	    // 9.5 Cleanup the view definition.
	    rewrite.clean();

	    rewritingsV.addElement(rewrite);

	}

	return rewritingsV;
    } 

  /**
  * This function will generate all the derived view definitions including
  * basic view definitions from the Vector of the basic view definitions.
  * @param rewritingsV basic legal rewritings of view definition.
  * @return derived legal rewritings of view definition.
  */
  Vector computeDerived(Vector rewritingsV){
    
  	Vector result = new Vector();

	  Enumeration enum = rewritingsV.elements();

  	while(enum.hasMoreElements()){
  	    // For each view in the rewritingsV.
  	    View_Definition v = (View_Definition)enum.nextElement();
        
  	    // Get all the components that can be dropped.
  	    Vector dropset = (Vector)v.getDispensableComponent();
  
	      if(EVEDemo.MW.Main.DEBUG)
    	  	System.out.println("Debug: View is: " + v);

  	    if(EVEDemo.MW.Main.DEBUG)
	      	System.out.println("Debug: VSA.computeDerived(): Dropset is: " + dropset);

  	    // Generate subsets from the Vector.
	      Subset set = new Subset(dropset);

	      Enumeration eset = set.getSubsets().elements();
  
	      while(eset.hasMoreElements()){
      		// get each subset
		      Vector s = (Vector)eset.nextElement();

                      if(EVEDemo.MW.Main.DEBUG)
                          System.out.println("Debug: VSA.computeDerived(): Subset is: " + s);
                      
                      // clone the view
                      View_Definition sv = (View_Definition)v.clone();
                      
                      // generate derived view
                      sv.dropComponents(s);
                      
                      result.addElement(sv);
	      }
  	}
        
	  return result;
  }

  /**
  * Return union of two EVEVectors of View_Definitions.
  */
    public static Vector union(Vector v1, Vector v2){

	if(v1 == null)
	    return deepClone(v2);

	if(v2 == null)
	    return deepClone(v1);

	Vector result = deepClone(v1);
  
	for(int i = 0; i<v2.size(); i++)
	    if(!result.contains(v2.elementAt(i)))
		result.addElement(v2.elementAt(i));
    
	return result;

    }

    /**
  * DeepClone a Vector;
  */
    public static Vector deepClone(Vector v){

	if(v == null)
	    return null;

	Vector result = new Vector();

	for(int i = 0; i< v.size(); i++){
	    Object obj = v.elementAt(i);
	    if(obj instanceof Vector)
		result.addElement(deepClone((Vector)obj));
	    else
		try{
		    result.addElement(obj.getClass().getDeclaredMethod("clone",null).invoke(obj, null));
		}catch(NoSuchMethodException nsme){
		    result.addElement(obj);
		    //System.out.println("VSA:deepClone one element cannot be cloned: " + nsme.getMessage());
		}catch(IllegalAccessException iae){
		    result.addElement(obj);
		    //System.out.println("VSA:deepClone one element cannot be clones: " + iae.getMessage());
		}catch(InvocationTargetException ite){
		    result.addElement(obj);
		    //System.out.println("VSA:deepClone one element cannot be clones: " + ite.getMessage());
		}
	}

	return result;
    }

}
