////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package EVEDemo.MW;

import DSRG.MKB.*;
import java.util.*;
import DSRG.VKB.*;
import DSRG.MW.*;

/**
 * This holds views. Mainly used for the EVE-Demo. It holds a view
 * extent, a view definition, and can hold another instance of this
 * same class (as an old view; used for the demo to show "extent
 * differences" ).
 * @author koeller,kajal */

public class View {

    private static final boolean DEBUG = false;

    /**
     * a handle to the Main class. This is used so the View class can
     * find the QueryEngine */
    private Main main;

  /**
   * name of the view
   */
  private String name;

  /**
   * view definition in E-SQL
   */
  public String esql_definition;
  /**
   * view definition in SQL
   */
  private String sql_definition;

  /**
   *
   * each instance includes another instance of the same class
   * in order to hold one "old" version of itself. <BR>
   * I'm still amazed that this is not recursive...
   *
   */
  private View oldView;

    /** 
     * with the advent of the QC-model, each view has to be capable of
     * temporarily holding a number of "versions", i.e., objects of
     * class "Version" which are needed for selecting a good view
     * after a schema change with the QC-Model.  The Version-class
     * sits in the MW package.
     * @author koeller */
    
    private Vector versions;
  
  /**
   * the view can hold its own extent as a Table object. <BR>
   * this is not feasible when view extents are large, but this is
   * only a demo :-) */
  private Table table;
  
  /**
   * handle to the current PSWEEP instance <BR>
   * used for incremental VM
   */
  private PSWEEP.PSWEEP myPSWEEP;
  
  /**
   * handle to the data update receiver in SWEEP <BR>
   * used for performance experiments
   */
  private SWEEP.DataUpdateReceiver myDUR;
  
  /**
   * handle to the current VKB instance <BR> used for all kinds of
   * stuff of Jeff's algorithm. All the view instance must use same
   * VKB object.  */
  private static DSRG.VKB.VKB mainVKB;


  /**
   * 0 for WHERE clause
   */
  public static final int WHERE_CLAUSE  = 0;
  /**
   * 1 for SELECT clause
   */
  public static final int SELECT_CLAUSE = 1;
  /**
   * 2 for FROM clause
   */
  public static final int FROM_CLAUSE    = 2;

  /**
   * has there been a change made? It's for time measurement of
   * performance evaluation
   */
  private boolean changed = false;

  /**
   * 0 for LOCAL clause
   */
  public static final int LOCAL_WHERE  = 0;
  /**
   * 1 for INTERNAL clause
   */
  public static final int INTERNAL_WHERE = 1;
  /**
   * 2 for EXTERNAL clause
   */
  public static final int EXTERNAL_WHERE = 2;
  
  /**
   * timestamp for performance experiments
   */
  long startPerformance;


  /**
   * Constructs a new instance of the View class <BR>
   */
    public View (Main myMain, String vName, 
		DSRG.MKB.MKB mainMKB,
		DSRG.VKB.VKB myVKB){
        main = myMain;
    name = vName;
    //ISs = mainMKB.getISs();
    oldView = null;
    
    if(myVKB != null){
      mainVKB = myVKB;
    }
    
    //ISNames = mainMKB.getAllISsAsStrings();

    // Using the name get the view definition (esql and sql) from the VKB

    esql_definition = mainVKB.getESQL ( name );
    sql_definition = mainVKB.getSQL ( name );

    /*
      System.out.println("============================================");
      System.out.println("============================================");
      long currenttime = System.currentTimeMillis();
      System.out.println("TIME before: " + currenttime);
      System.out.println("============================================");
      System.out.println("============================================");
    */
    table = main.getQueryEngine().computeExtent(this);
    /*
      System.out.println("============================================");
      System.out.println("============================================");
      System.out.println("TIME after: "+System.currentTimeMillis());
      System.out.println("TIME passed: "+(System.currentTimeMillis()-
      currenttime)/1000.0+" s.");
      System.out.println("============================================");
      System.out.println("============================================");
    */
    
    
    try{
      myPSWEEP = new PSWEEP.PSWEEP(this, mainMKB);
      //myPSWEEP.start();
      myDUR = myPSWEEP.getReceiver();
    }
    catch (Exception e) {
      System.out.println("PSWEEP Exception\n");
      e.printStackTrace();
      System.out.println(e.getMessage());
    }
    // add support for QC-Model.
    versions = new Vector(); // flush versions vector;
    versions = ((View_Definition)mainVKB.getVD(name)).getVersions();
    
  }
    
  /**
   * Copy Constructor. Makes a copy of the view to a second identical
   * variable structure in this instance. <BR> For comparing old and
   * new view extents in the demo */
    public View ( View oldView ){
	this.name = oldView.name;
	this.esql_definition = oldView.esql_definition;
	this.sql_definition = oldView.sql_definition;
	this.oldView = null;
        this.table = new Table(oldView.table);
        //	this.ISs = oldView.ISs;
	this.mainVKB = oldView.mainVKB;
	if(mainVKB == null)
	    System.out.println("Debug: View.View(View) VKB is null!");
    }
    
    /**
   * getName - returns the name of the view
   */
  public String getName ( ){
    return name;
  }

  /**
   * Refresh - refresh the current view from the VKB.  <BR> Hacked
   * 030799 by koeller. This function should (ideally) be called only
   * after a view synchronization. So we can retrieve a vector of
   * versions from the VKB so the GUI can query it.  */

    public void refresh(){
	// refresh ESQL from VKB
	esql_definition = mainVKB.getESQL( name );
	
	String old_sql_def = sql_definition;
	// refresh SQL from VKB
	sql_definition = mainVKB.getSQL( name );
        
	if (DEBUG)
	    System.out.println("View.refresh() old_sql:" + 
                               old_sql_def + 
                               "\nView.refresh() new_sql:" + 
                               sql_definition);
	
	if(!old_sql_def.equals(sql_definition)){
	    
            table = main.getQueryEngine().computeExtent(this);
            if (DEBUG) System.out.println(table);
	}
        

        // add support for QC-Model. QCMODEL
        versions = new Vector(); // flush versions vector;
        View_Definition vd = mainVKB.getVD(name);
        if (vd!=null)
	    versions = vd.getVersions();
        if (DEBUG) 
            System.out.println("Versions from View.refresh: "+versions);
    }
    
    
    /**
     * getDefinition - returns the definition of this view.
     * This defaults to returning the esql definition since that is what is
     * needed by the GUI.
     */
    public String getDefinition ( ){
        return esql_definition;
    }
    
    /** get the extent as the new table object */
    public Table getExtent() {
        return table;
    }
    
    /**
     * Set the new table of this View. Needed for incremental Updates.
     */
    public void setExtent(Table ext){
        table = ext;
    }
    
    /**
     * getOldView - returns a reference to the OldView.
     */
    public View getOldView ( ){
        return oldView;
    }
    
  /**
   * saveView - makes a copy of the view into the oldView.
   * if oldView already exists then, it overrides it.
   */
    public void saveView ( ){
    // call the copy constructor
        oldView = new View ( this );
    }
    
    
    //===================== For Performance Evaluation =====
    // created by Andreas
    
  /**
   * Stop the timer for the Data Update performance evaluation.
   */
  public void stopTimer(){
    changed = true; // doesn't work due to unmarshal exception, hence the refresh button
    System.out.println("============================================");
    System.out.println("============================================");
    System.out.println("TIME after: "+System.currentTimeMillis());
    System.out.println("TIME passed: "+(System.currentTimeMillis()-
					startPerformance)/1000.0+" s.");
    System.out.println("============================================");
    System.out.println("============================================");
  }

  public boolean getChangeStatus() {
    
    if(changed)
      {
	changed = false;
	return true;
      }
    else
      return false;
  }

  //=========== End of Performance Evaluation ===================

  //================ begin of the SWEEP integration ===================
  // functions added by Andreas depends on Xin's requirement of SWEEP

  /**
   * This is used to send data update to the information source.
   * It's used for SWEEP
   */
  public void sendUpdate(String isName, String relName, int count, String update) {
   myDUR.putDataUpdate(isName, relName, count, update);
  }

    /**
     * Used to start the View Maintenance Module
     */
    public void startVM(){
	myPSWEEP.start();
    }


    /** if we have negative counter, i.e. fail to delete something, we
    * store those in this relation.  then, after each addition, we try
    * to apply this delete relation again to view, if that is
    * successful, then, we remove the tuple from it, otherwise, we
    * still keep the tuple.*/
    SWEEP.Relation negRel = null;

  /** 
   *   the interface with the SWEEP classes <BR>
   * adds the incoming SWEEP.Relation-object to the extent
   * <BR>DISABLED FOR RE-IMPLEMENTATION
   */

  public void addTuples(SWEEP.Relation sweepRelation){
      System.err.println("called View::addTuples , but this is disabled");
  }  
    /*
  public void addTuples(SWEEP.Relation sweepRelation){
    SWEEP.Table relExtent = sweepRelation.getExtent();
    SWEEP.Tuple tuple;
    String field;
    Attribute attr=null;
    int i,j,k;
    System.out.println("Adding Tuples to extent");

    try {
      for (i = 0 ; i < relExtent.size(); i++) {
        tuple = relExtent.getTuple(i);
        SWEEP.Tuple t;
        // need to check whether the tuple can be found in the negRel
        if(negRel != null){

          if(EVEDemo.MW.Main.DEBUG)
      	    System.out.println("DEBUG: View.addTuples checking tuple: " + tuple );
      
      	  if(( t = negRel.getExtent().find(tuple))!=null){// if we found that tuple in negative relation
            if(EVEDemo.MW.Main.DEBUG)
        	  	System.out.println("DEBUG: View.addTuples found the tuple in negRel (t: " + t + ")");
      
      		// reset the tuple's counter
      		tuple.setNumber(tuple.getNumber() + t.getNumber());
      		
      		if(tuple.getNumber() >= 0){// if still has something to add, or nothing left
      		    // remove the tuple from the negRel.
      		    t.setNumber(0);
      
      		}else{//tuple.getNumber()< 0, still has negative part
      		    // descrease the negative counter of that tuple.
      		    t.setNumber(tuple.getNumber());
      		}
        }else{// we haven't found that tuple in the negRel, do nothing.
	    }
	}//if we have negRel

	// insert the tuple into EVE view., here if the
	// tuple.getNumber < 0, we still do nothing. It's a trick!!!
	for(k=0; k < tuple.getNumber(); k++){
	    for (j = 0 ; j < table.getRelation().getAttributes().size(); j++) {
		field = (String)tuple.getField(j);
		attr = (Attribute)table.getRelation().getAttributes().elementAt(j);
		attr.getDataValues().addElement(new DataValue(attr, field));
	    }
	}

      }// end for each tuple
    }
    
    catch (Exception e) {
      e.printStackTrace();
    }
    if (attr!=null) 
      System.out.println("View::addTuples() size of attributes:"+attr.getDataValues().size());
    else System.out.println("View::addTuples()  Attribute is null.");
  }
  
    */


  /** the interface with the SWEEP classes <BR>
   * removes the incoming SWEEP.Relation-object from the extent
   * <BR><B>DISABLED FOR RE-IMPLEMENTATION</B>
   */
  
  public void removeTuples(SWEEP.Relation sweepRelation){
      System.err.println("called View::addTuples , but this is disabled");
  }  

    /*
  public void removeTuples(SWEEP.Relation sweepRelation){

    SWEEP.RelationExtent relExtent = sweepRelation.getExtent();
    SWEEP.Tuple tuple;
    String field, value;
    Vector attributes = table.getRelation().getAttributes();
    Attribute attr= null;
    DataValue dataValue;
    int tupleIndex, tupleCount, fieldIndex, lineNumber,
      columnCount = table.getRelation().getAttributes().size(),
      extentSize = ((Attribute)table.getRelation().getAttributes().elementAt(0)).getDataValues().size();
    boolean tuple_removed;
  
    if(EVEDemo.MW.Main.DEBUG)
      System.out.println("Removing tuples from extent");
    try {
      // for all tuples in the delta relation
      for (tupleIndex = 0 ; tupleIndex < relExtent.size(); tupleIndex++) {
	// take the tuple
	tuple = relExtent.getTuple(tupleIndex);
	// and the multiplicity
	for(tupleCount=0; tupleCount < -tuple.getNumber(); tupleCount++){ // tuple counts are negative here...
	  lineNumber = -1;
	  fieldIndex = 0;
	  tuple_removed = false;
	  // search extent from the first line
	  do {
	    lineNumber++;
	    // get a field
	    field = new String((String)tuple.getField(fieldIndex));
	    attr = (Attribute)attributes.elementAt(fieldIndex);
	    dataValue = (DataValue)attr.getDataValues().elementAt(lineNumber);
	    value = new String(dataValue.getValue());
	    // and compare it with the delta tuple field
	    while (value.equals(field) && fieldIndex < columnCount-1) {
	      // if one matches, check the other columns one by one
	      fieldIndex++;
	      field = new String((String)tuple.getField(fieldIndex));
	      attr = (Attribute)attributes.elementAt(fieldIndex);
	      dataValue = (DataValue)attr.getDataValues().elementAt(lineNumber);
	      value = new String(dataValue.getValue());
	    }
	    // if all fields match
	    if (fieldIndex == columnCount-1 && value.equals(field)) { // success
	      // remove tuple
	      tuple_removed = true;
	      extentSize--;
	      if(EVEDemo.MW.Main.DEBUG)
          System.out.println("One tuple removed");
	      // i.e., remove a value from all attributes
	      for (fieldIndex = 0; fieldIndex < columnCount; fieldIndex++) {
		attr = (Attribute)attributes.elementAt(fieldIndex);
		attr.getDataValues().removeElementAt(lineNumber);
	      }
	    }
	    // otherwise, continue with next line in extent, first field
	    fieldIndex = 0;
	    // until all extent tuples are searched through
	  } while(!tuple_removed && lineNumber < extentSize-1);
    
	  //if (!tuple_removed && lineNumber == extentSize)
	  if(!tuple_removed){// modified by xinz.
	      System.out.println("Tuple to remove not found");
	      if(negRel == null){
		  // clone the structure, but erase the extent.
		  negRel = (SWEEP.Relation)sweepRelation.clone();
		  negRel.setExtent(new SWEEP.RelationExtent());
	      }
	      // add the unremoved tuple to this relation
	      SWEEP.Tuple t = (SWEEP.Tuple)tuple.clone();
	      t.setNumber(-1);
	      negRel.insertTuple(t);
        if(EVEDemo.MW.Main.DEBUG)
	        System.out.println("The faulty tuple is added to the negRel (" + negRel.getExtent().size()+ ").");
        if(EVEDemo.MW.Main.DEBUG)
	        System.out.println("The negRel is: " + negRel);
	  }
	} // next tuple
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    if (attr!=null) 
      System.out.println("View::removeTuples() size of attributes:"+attr.getDataValues().size());
    else System.out.println(" View::removeTuples()  Attribute is null.");
  }
    */  

  //====================== end of SWEEP integration ===============


    // =============================================================
    // =============================================================
    // =============================================================
    // =============================================================
    // ==============    QC-Model Integration    ===================
    // =============================================================
    // =============================================================
    // =============================================================
    // =============================================================

    /**
     * returns the view definition string of the version with index "index"
     * as a string
     * @param index the index of the version to return
     * @return null if index is wrong
     */
    public String getVersion(int index) {
	if (versions==null) return null;

	if (index>=0 && index < versions.size() ){
	    return ((Version)versions.elementAt(index)).getDefinition();
	}
	else return null;
    }

    /** calls the VKB to replace this view's View_Definition by
        version number i */
    public boolean replaceView_DefinitionBy(int version_number) {
        if (DEBUG) System.out.println("View.replaceView_DefinitionBy: we have "+versions.size()+"versions");
        if (DEBUG) System.out.println("View.replaceView_DefinitionBy: replacing View by version "+version_number);
        if (version_number < 0 || version_number >= versions.size())
            return false;
        Version tmpVersion = (Version)versions.elementAt(version_number);
        mainVKB.replaceViewDefinition(tmpVersion.getOldDefinitionAsObject(),
                                      tmpVersion.getDefinitionAsObject());
        // kill the versions (for the next VS)
        tmpVersion.getOldDefinitionAsObject().setVersions(null);
        refresh(); // not sure if this is good coding
        return true;
    }
    

    /**
     * returns the number of versions for this view or null if the
     * vector is not intialized
     * @return -1 is the versions vector is not initialized */
    public int numberOfVersions() {
	if (DEBUG) System.out.println("View.numberOfVersions(): Versions Vector is "+versions);
	if (versions!=null)
	    return versions.size();
	else return -1;
    }

    /**
     * returns the QC-Object for the version at "index"
     * @param index the index of the version for which to return the
     * QC-Object 
     * @return null if index is wrong */
    
    public QC getQC(int index) {
	if (versions==null) return null;

	if (index>=0 && index < versions.size() )
	    return ((Version)versions.elementAt(index)).getQC();
	else return null;
    }
    
    
}
